# -*- coding: utf-8 -*-
import os
import fnmatch
import logging

# third-party
import requests
from crosspm.helpers.config import CROSSPM_DEPENDENCY_LOCK_FILENAME
from crosspm.helpers.exceptions import *


# log = logging.getLogger(__name__)


# TODO: REWORK PROMOTER COMPLETELY to match new architecture!!!
class Promoter:
    def __init__(self, config, depslock_path=''):
        self._log = logging.getLogger('crosspm')
        self._config = config
        self._cache_path = config.crosspm_cache_root
        if not os.path.exists(self._cache_path):
            os.makedirs(self._cache_path)

        self.packed_path = os.path.realpath(os.path.join(self._cache_path, 'archive'))
        self.unpacked_path = os.path.realpath(os.path.join(self._cache_path, 'cache'))
        self.temp_path = os.path.realpath(os.path.join(self._cache_path, 'tmp'))

        if not depslock_path:
            depslock_path = config.deps_lock_file_name \
                if config.deps_lock_file_name \
                else CROSSPM_DEPENDENCY_LOCK_FILENAME
        self._depslock_path = os.path.realpath(depslock_path)

    @staticmethod
    def get_version_int(version_str):
        parts = version_str.split('.')
        try:
            parts = list(map(int, parts))
            if len(parts) < 4:
                parts.insert(2, 0)
            return parts

        except ValueError:
            return [0] * 4

        except IndexError:
            return [0] * 4

    def parse_dir_list(self, data_dict, dir_list_data):
        for item in dir_list_data["files"]:
            if item["folder"]:
                continue

            parts = item["uri"].split('/')

            name = parts[1]
            branch = parts[2]
            version = parts[3]
            version_int = self.get_version_int(version)

            data_dict.setdefault(name, {})
            data_dict[name].setdefault(branch, [])
            data_dict[name][branch].append((version, version_int))

    @staticmethod
    def join_package_path(server_url, part_a, repo, path):
        server_url = server_url if not server_url.endswith('/') else server_url[: -1]
        part_a = part_a if not part_a.startswith('/') else part_a[1:]
        part_a = part_a if not part_a.endswith('/') else part_a[: -1]
        repo = repo if not repo.startswith('/') else repo[1:]
        repo = repo if not repo.endswith('/') else repo[: -1]
        path = path if not path.startswith('/') else path[1:]

        return '{server_url}/{part_a}/{repo}/{path}'.format(**locals())

    def promote_packages(self, list_or_file_path=None):
        if list_or_file_path is None:
            list_or_file_path = self._depslock_path

        data_tree = {}
        deps_list = []  # pm_common.get_dependencies(os.path.join(os.getcwd(), CROSSPM_DEPENDENCY_FILENAME))
        out_file_data_str = ''
        out_file_format = '{:20s} {:20s} {}\n'
        out_file_path = os.path.join(os.getcwd(), CROSSPM_DEPENDENCY_LOCK_FILENAME)

        # clean file
        with open(out_file_path, 'w') as out_f:
            out_f.write(out_file_format.format('# package', '# version', '# branch\n'))

        for i, _src in enumerate(self._config.sources()):
            if i > 0:
                self._log.warning('')
                self._log.warning('Next source ...')

            for current_repo in _src['repo']:
                request_url = self.join_package_path(
                    _src['server'],
                    'api/storage',
                    current_repo,
                    '?list&deep=1&listFolders=0&mdTimestamps=1&includeRootPath=1',
                )

                self._log.info('GET request: %s', request_url)

                r = requests.get(
                    request_url,
                    verify=False,
                    auth=tuple(_src['auth']),
                )

                r.raise_for_status()

                data_response = r.json()

                self.parse_dir_list(data_tree, data_response)

        for (d_name, d_version, d_branch,) in deps_list:
            if d_name not in data_tree:
                raise CrosspmException(
                    CROSSPM_ERRORCODE_PACKAGE_NOT_FOUND,
                    'unknown package [{}]'.format(d_name),
                )

            if d_branch not in data_tree[d_name]:
                raise CrosspmException(
                    CROSSPM_ERRORCODE_PACKAGE_BRANCH_NOT_FOUND,
                    'unknown branch [{}] of package [{}]'.format(
                        d_branch,
                        d_name,
                    ))

            libs = data_tree[d_name][d_branch]

            versions = [v for v in libs if fnmatch.fnmatch(v[0], '.'.join(d_version))]

            if not versions:
                raise CrosspmException(
                    CROSSPM_ERRORCODE_VERSION_PATTERN_NOT_MATCH,
                    'pattern [{ver}] does not match any version of package [{pkg}] for branch [{br}] '.format(
                        ver='.'.join(d_version),
                        pkg=d_name,
                        br=d_branch,
                    ))

            current_version = max(versions, key=lambda x: x[1])

            out_file_data_str += out_file_format.format(
                d_name,
                current_version[0],
                d_branch
            )

        with open(out_file_path, 'w') as out_f:
            out_f.write(out_file_data_str)



# -*- coding: utf-8 -*-

from crosspm.helpers.locker import Locker


class Usedby(Locker):
    def __init__(self, config, do_load, recursive):
        # Ignore do_load flag
        super(Usedby, self).__init__(config, False, recursive)

    def usedby_packages(self, deps_file_path=None, depslock_file_path=None, packages=None):
        """
        Lock packages. Downloader search packages
        """
        if deps_file_path is None:
            deps_file_path = self._deps_path
        if depslock_file_path is None:
            depslock_file_path = self._depslock_path
        if deps_file_path == depslock_file_path:
            depslock_file_path += '.lock'

        if packages is None:
            self.search_dependencies(deps_file_path)
        else:
            self._root_package.packages = packages
        self._log.info('Done!')

    def search_dependencies(self, depslock_file_path):
        self._log.info('Check dependencies ...')
        self._root_package.find_usedby(depslock_file_path, property_validate=True)
        self._log.info('')
        self._log.info('Dependency tree:')
        self._root_package.print(0, self._config.output('tree', [{self._config.name_column: 0}]))

    def entrypoint(self, *args, **kwargs):
        self.usedby_packages(*args, **kwargs)



# -*- coding: utf-8 -*-
import contextlib
import os
import shutil
import sys
import tarfile
import zipfile

from crosspm.helpers.exceptions import CrosspmException, CROSSPM_ERRORCODE_NO_FILES_TO_PACK, \
    CROSSPM_ERRORCODE_UNKNOWN_ARCHIVE

WINDOWS = True if sys.platform == 'win32' else False


class Archive:
    @staticmethod
    def create(archive_name, src_dir_path):
        archive_name = os.path.realpath(archive_name)
        src_dir_path = os.path.realpath(src_dir_path)
        files_to_pack = []
        archive_name_tmp = os.path.join(
            os.path.dirname(archive_name),
            'tmp_{}'.format(os.path.basename(archive_name)),
        )

        for name in os.listdir(src_dir_path):
            current_path = os.path.join(src_dir_path, name)
            real_path = os.path.realpath(current_path)
            rel_path = os.path.relpath(current_path, src_dir_path)

            files_to_pack.append((real_path, rel_path,))

        if not files_to_pack:
            raise CrosspmException(
                CROSSPM_ERRORCODE_NO_FILES_TO_PACK,
                'no files to pack, directory [{}] is empty'.format(
                    src_dir_path
                ),
            )

        with contextlib.closing(tarfile.TarFile.open(archive_name_tmp, 'w:gz')) as tf:
            for real_path, rel_path in files_to_pack:
                tf.add(real_path, arcname=rel_path)

        os.renames(archive_name_tmp, archive_name)

    @staticmethod
    def extract(archive_name, dst_dir_path, file_name=None):
        # HACK for https://bugs.python.org/issue18199
        if WINDOWS:
            dst_dir_path = "\\\\?\\" + dst_dir_path

        dst_dir_path_tmp = '{}_tmp'.format(dst_dir_path)

        # remove temp dir
        if os.path.exists(dst_dir_path_tmp):
            shutil.rmtree(dst_dir_path_tmp)
        if os.path.exists(dst_dir_path):
            os.renames(dst_dir_path, dst_dir_path_tmp)

        # TODO: remove hack for deb-package
        # ------- START deb-package -------
        if archive_name.endswith(".deb"):
            # На Windows:
            # deb-пакет содержит в себе data.tar файл
            # Распаковываем deb в первый раз, сохраняя data.tar рядом с deb-файлом
            # Затем, подменяет имя архива и пускаем по обычному пути распаковки,
            # чтобы он нашел tar и распаковал его как обычно
            if WINDOWS:
                datatar_dir = os.path.dirname(archive_name)
                if not os.path.exists(datatar_dir):
                    os.makedirs(datatar_dir)
                from pyunpack import Archive
                Archive(archive_name).extractall(datatar_dir)
                # Подмена имени
                archive_name = os.path.join(datatar_dir, 'data.tar')
            # На linux - распаковываем сразу
            else:
                from pyunpack import Archive
                if not os.path.exists(dst_dir_path):
                    os.makedirs(dst_dir_path)
                Archive(archive_name).extractall(dst_dir_path)
        # ------- END deb-package -------

        if tarfile.is_tarfile(archive_name):
            with contextlib.closing(tarfile.TarFile.open(archive_name, 'r:*')) as tf:
                if file_name:
                    tf.extract(file_name, path=dst_dir_path)
                else:
                    # Quick hack for unpack tgz on Windows with PATH like \\?\C:
                    # Internal issue: CM-39214
                    try:
                        tf.extractall(path=dst_dir_path)
                    except OSError:
                        dst_dir_path = dst_dir_path.replace("\\\\?\\", '')
                        tf.extractall(path=dst_dir_path)

        elif zipfile.is_zipfile(archive_name):
            with contextlib.closing(zipfile.ZipFile(archive_name, mode='r')) as zf:
                if file_name:
                    zf.extract(file_name, path=dst_dir_path)
                else:
                    zf.extractall(path=dst_dir_path)

        elif archive_name.endswith('.rar'):
            if not os.path.exists(dst_dir_path):
                os.makedirs(dst_dir_path)
            from pyunpack import Archive
            Archive(archive_name).extractall(dst_dir_path)

        elif archive_name.endswith('.deb'):
            # We unpack above
            pass

        else:
            tries = 0
            while tries < 3:
                tries += 1
                try:
                    if os.path.exists(dst_dir_path):
                        shutil.rmtree(dst_dir_path)
                    if os.path.exists(dst_dir_path_tmp):
                        os.renames(dst_dir_path_tmp, dst_dir_path)
                    tries = 3
                except Exception:
                    pass
            raise CrosspmException(
                CROSSPM_ERRORCODE_UNKNOWN_ARCHIVE,
                'unknown archive type. File: [{}]'.format(archive_name),
            )

        # remove temp dir
        if os.path.exists(dst_dir_path_tmp):
            shutil.rmtree(dst_dir_path_tmp)

    @staticmethod
    def extract_file(archive_name, dst_dir_path, file_name):

        try:
            Archive.extract(archive_name, dst_dir_path, file_name)
            _dest_file = os.path.join(dst_dir_path, file_name)
        except Exception:
            _dest_file = None
        return _dest_file



# -*- coding: utf-8 -*-
import logging


class BaseAdapter:
    def __init__(self, config):
        self._config = config
        self._log = logging.getLogger('crosspm')

    def get_packages(self, source, parser, downloader, list_or_file_path, property_validate=True):
        # Here must be function
        self._log.info('Get packages')
        return None

    def download_package(self, package, dest_path):
        # Here must be function
        self._log.info('Download package')
        return None, False



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
{app_name}
Usage:
    crosspm download [options]
    crosspm lock [DEPS] [DEPSLOCK] [options]
    crosspm usedby [DEPS] [options]
    crosspm pack <OUT> <SOURCE> [options]
    crosspm cache [size | age | clear [hard]]
    crosspm -h | --help
    crosspm --version

Options:
    <OUT>                                Output file.
    <SOURCE>                             Source directory path.
    -h, --help                           Show this screen.
    --version                            Show version.
    -L, --list                           Do not load packages and its dependencies. Just show what's found.
    -v LEVEL, --verbose=LEVEL            Set output verbosity: ({verb_level}) [default: ].
    -l LOGFILE, --log=LOGFILE            File name for log output. Log level is '{log_default}' if set when verbose doesn't.
    -c FILE, --config=FILE               Path to configuration file.
    -o OPTIONS, --options OPTIONS        Extra options.
    --deps-path=FILE                     Path to file with dependencies [./{deps_default}]
    --depslock-path=FILE                 Path to file with locked dependencies [./{deps_lock_default}]
    --dependencies-content=CONTENT       Content for dependencies.txt file
    --dependencies-lock-content=CONTENT  Content for dependencies.txt.lock file
    --lock-on-success                    Save file with locked dependencies next to original one if download succeeds
    --out-format=TYPE                    Output data format. Available formats:({out_format}) [default: {out_format_default}]
    --output=FILE                        Output file name (required if --out_format is not stdout)
    --output-template=FILE               Template path, e.g. nuget.packages.config.j2 (required if --out_format=jinja)
    --no-fails                           Ignore fails config if possible.
    --recursive=VALUE                    Process all packages recursively to find and lock all dependencies
    --prefer-local                       Do not search package if exist in cache
    --stdout                             Print info and debug message to STDOUT, error to STDERR. Otherwise - all messages to STDERR

"""  # noqa

import logging
import os
import shlex
import sys
import time

from docopt import docopt

from crosspm import version
from crosspm.helpers.archive import Archive
from crosspm.helpers.config import (
    CROSSPM_DEPENDENCY_LOCK_FILENAME,
    CROSSPM_DEPENDENCY_FILENAME,
    Config,
)
from crosspm.helpers.content import DependenciesContent
from crosspm.helpers.downloader import Downloader
from crosspm.helpers.exceptions import *  # noqa
from crosspm.helpers.locker import Locker
from crosspm.helpers.output import Output
from crosspm.helpers.python import get_object_from_string
from crosspm.helpers.usedby import Usedby

app_name = 'CrossPM (Cross Package Manager) version: {version} The MIT License (MIT)'.format(version=version)


def do_run(func):
    def wrapper(self, *args, **kwargs):
        try:
            res = func(self, *args, **kwargs)
        except CrosspmExceptionWrongArgs as e:
            print(__doc__)
            return self.exit(e.error_code, e.msg)

        except CrosspmException as e:
            return self.exit(e.error_code, e.msg)

        except Exception as e:
            self._log.exception(e)
            return self.exit(CROSSPM_ERRORCODE_UNKNOWN_ERROR, 'Unknown error occurred!')
        return 0, res

    return wrapper


class CrossPM:
    _ready = False

    def __init__(self, args=None, throw_exceptions=None, return_result=False):
        self._config = None
        self._output = None
        self._return_result = return_result

        if throw_exceptions is None:
            # legacy behavior
            if self._return_result:
                self._throw_exceptions = False
            else:
                self._throw_exceptions = True
        else:
            self._throw_exceptions = throw_exceptions

        self._log = logging.getLogger('crosspm')

        args = self.prepare_args(args)
        docopt_str = __doc__.format(app_name=app_name,
                                    verb_level=Config.get_verbosity_level(),
                                    log_default=Config.get_verbosity_level(0, True),
                                    deps_default=CROSSPM_DEPENDENCY_FILENAME,
                                    deps_lock_default=CROSSPM_DEPENDENCY_LOCK_FILENAME,
                                    out_format=Output.get_output_types(),
                                    out_format_default='stdout',
                                    )
        self._args = docopt(docopt_str,
                            argv=args,
                            version=version)

        if self._args['--recursive']:
            recursive_str = self._args['--recursive']
            if recursive_str.lower() == 'true':
                self._args['--recursive'] = True
            elif recursive_str.lower() == 'false':
                self._args['--recursive'] = False
            else:
                raise Exception("Unknown value to --recursive: {}".format(recursive_str))

        if isinstance(self._args, str):
            if self._throw_exceptions:
                print(app_name)
                print(self._args)
                exit()

        self._ready = True

        if self._args['download']:
            self.command_ = Downloader
        elif self._args['lock']:
            self.command_ = Locker
        elif self._args['usedby']:
            self.command_ = Usedby
        else:
            self.command_ = None

    @property
    def stdout(self):
        """
        Флаг --stdout может быть взят из переменной окружения CROSSPM_STDOUT.
        Если есть любое значение в CROSSPM_STDOUT - оно понимается как True
        :return:
        """
        # --stdout
        stdout = self._args['--stdout']
        if stdout:
            return True

        # CROSSPM_STDOUT
        stdout_env = os.getenv('CROSSPM_STDOUT', None)
        if stdout_env is not None:
            return True

        return False

    @staticmethod
    def prepare_args(args, windows=None):
        """
        Prepare args - add support for old interface, e.g:
            - --recursive was "flag" and for now it support True or False value
        :param args:
        :return:
        """
        if windows is None:
            windows = "win" in sys.platform

        if isinstance(args, str):
            args = shlex.split(args, posix=not windows)
        elif isinstance(args, list):
            pass
        elif args is None:
            args = sys.argv[1:]
        else:
            raise Exception("Unknown args type: {}".format(type(args)))

        # --recursive => --recursive=True|False convert
        for position, argument in enumerate(args):
            # Normal way, skip change
            if argument.lower() in ('--recursive=true', '--recursive=false'):
                return args

            elif argument.lower() == '--recursive':
                if len(args) > position + 1 and args[position + 1].lower() in ["true", "false"]:
                    # --recursive true | false
                    return args
                else:
                    # legacy way, convert --recursive to --recursive=true
                    args[position] = "--recursive=True"
                    return args
        return args

    @do_run
    def read_config(self):
        _deps_path = self._args['--deps-path']
        # Передаём содержимое напрямую
        if _deps_path is None and self._args['--dependencies-content'] is not None:
            _deps_path = DependenciesContent(self._args['--dependencies-content'])
        _depslock_path = self._args['--depslock-path']
        if _depslock_path is None and self._args['--dependencies-lock-content'] is not None:
            _depslock_path = DependenciesContent(self._args['--dependencies-lock-content'])
        if self._args['lock']:
            if self._args['DEPS']:
                _deps_path = self._args['DEPS']
            if self._args['DEPSLOCK']:
                _depslock_path = self._args['DEPSLOCK']
        self._config = Config(self._args['--config'], self._args['--options'], self._args['--no-fails'], _depslock_path,
                              _deps_path, self._args['--lock-on-success'],
                              self._args['--prefer-local'])
        self._output = Output(self._config.output('result', None), self._config.name_column, self._config)

    def exit(self, code, msg):
        self._log.critical(msg)
        if self._throw_exceptions:
            sys.exit(code)
        else:
            return code, msg

    @property
    def recursive(self):
        if self.command_ is Downloader:
            if self._args['--recursive'] is None:
                recursive = True
            else:
                recursive = self._args['--recursive']
        else:
            if self._args['--recursive'] is None:
                recursive = False
            else:
                recursive = self._args['--recursive']
        return recursive

    @do_run
    def check_common_args(self):
        if self._args['--output']:
            output = self._args['--output'].strip().strip("'").strip('"')
            output_abs = os.path.abspath(output)
            if os.path.isdir(output_abs):
                raise CrosspmExceptionWrongArgs(
                    '"%s" is a directory - can\'t write to it'
                )
            self._args['--output'] = output

    @do_run
    def set_logging_level(self):
        level_str = self._args['--verbose'].strip().lower()

        log = self._args['--log']
        if log:
            log = log.strip().strip("'").strip('"')
            log_abs = os.path.abspath(log)
            if os.path.isdir(log_abs):
                raise CrosspmExceptionWrongArgs(
                    '"%s" is a directory - can\'t write log to it'
                )
            else:
                log_dir = os.path.dirname(log_abs)
                if not os.path.exists(log_dir):
                    os.makedirs(log_dir)
        else:
            log_abs = None

        level = Config.get_verbosity_level(level_str or 'console')
        self._log.handlers = []
        if level or log_abs:
            self._log.setLevel(level)
            format_str = '%(asctime)-19s [%(levelname)-9s] %(message)s'
            if level_str == 'debug':
                format_str = '%(asctime)-19s [%(levelname)-9s] %(name)-12s: %(message)s'
            formatter = logging.Formatter(format_str, datefmt="%Y-%m-%d %H:%M:%S")

            if level:
                # legacy way - Cmake catch message from stdout and parse PACKAGE_ROOT
                # So, crosspm print debug and info message to stderr for debug purpose
                if not self.stdout:
                    sh = logging.StreamHandler(stream=sys.stderr)
                    sh.setLevel(level)
                    self._log.addHandler(sh)
                # If --stdout flag enabled
                else:
                    sh = logging.StreamHandler(stream=sys.stderr)
                    sh.setLevel(logging.WARNING)
                    self._log.addHandler(sh)
                    sh = logging.StreamHandler(stream=sys.stdout)
                    sh.setLevel(level)
                    self._log.addHandler(sh)

            if log_abs:
                if not level_str:
                    level = Config.get_verbosity_level(0)
                fh = logging.FileHandler(filename=log_abs)
                fh.setLevel(level)
                fh.setFormatter(formatter)
                self._log.addHandler(fh)

    def run(self):
        time_start = time.time()
        if self._ready:

            errorcode, msg = self.set_logging_level()
            self._log.info(app_name)
            errorcode, msg = self.check_common_args()
            if errorcode == 0:
                errorcode, msg = self.read_config()

                if errorcode == 0:
                    if self._args['download']:
                        errorcode, msg = self.command(self.command_)
                    elif self._args['lock']:
                        errorcode, msg = self.command(self.command_)
                    elif self._args['usedby']:
                        errorcode, msg = self.command(self.command_)
                    elif self._args['pack']:
                        errorcode, msg = self.pack()
                    elif self._args['cache']:
                        errorcode, msg = self.cache()
        else:
            errorcode, msg = CROSSPM_ERRORCODE_WRONG_ARGS, self._args
        time_end = time.time()
        self._log.info('Done in %2.2f sec' % (time_end - time_start))
        return errorcode, msg

    @do_run
    def command(self, command_):
        if self._return_result:
            params = {}
        else:
            if self._args['--out-format'] == 'stdout':
                if self._args['--output']:
                    raise CrosspmExceptionWrongArgs(
                        "unwanted argument '--output' while argument '--out-format={}'".format(
                            self._args['--out-format'],
                        ))
            elif not self._args['--output']:
                raise CrosspmExceptionWrongArgs(
                    "argument '--output' required when argument '--out-format={}'".format(
                        self._args['--out-format'],
                    ))

            params = {
                'out_format': ['--out-format', ''],
                'output': ['--output', ''],
                'output_template': ['--output-template', ''],
                # 'out_prefix': ['--out-prefix', ''],
                # 'depslock_path': ['--depslock-path', ''],
            }

            for k, v in params.items():
                params[k] = self._args[v[0]] if v[0] in self._args else v[1]
                if isinstance(params[k], str):
                    params[k] = params[k].strip('"').strip("'")

            # try to dynamic load --output-template from python module
            output_template = params['output_template']
            if output_template:
                # Try to load from python module
                module_template = get_object_from_string(output_template)
                if module_template is not None:
                    self._log.debug(
                        "Found output template path '{}' from '{}'".format(module_template, output_template))
                    params['output_template'] = module_template
                else:
                    self._log.debug("Output template '{}' use like file path".format(output_template))

            # check template exist
            output_template = params['output_template']
            if output_template and not os.path.exists(output_template):
                raise CrosspmException(CROSSPM_ERRORCODE_CONFIG_NOT_FOUND,
                                       "Can not find template '{}'".format(output_template))

        do_load = not self._args['--list']
        # hack for Locker
        if command_ is Locker:
            do_load = self.recursive

        cpm_ = command_(self._config, do_load, self.recursive)
        cpm_.entrypoint()

        if self._return_result:
            return self._return(cpm_)
        else:
            # self._output.write(params, packages)
            self._output.write_output(params, cpm_.get_tree_packages())
        return ''

    def _return(self, cpm_downloader):
        if str(self._return_result).lower() == 'raw':
            return cpm_downloader.get_raw_packages()
        if str(self._return_result).lower() == 'tree':
            return cpm_downloader.get_tree_packages()
        else:
            return self._output.output_type_module(cpm_downloader.get_tree_packages())

    @do_run
    def pack(self):
        Archive.create(self._args['<OUT>'], self._args['<SOURCE>'])

    @do_run
    def cache(self):
        if self._args['clear']:
            self._config.cache.clear(self._args['hard'])
        elif self._args['size']:
            self._config.cache.size()
        elif self._args['age']:
            self._config.cache.age()
        else:
            self._config.cache.info()


if __name__ == '__main__':
    app = CrossPM()
    app.run()



# -*- coding: utf-8 -*-
import logging
import os
import time
# from crosspm.helpers.package import Package
from datetime import datetime, timedelta


class Cache:
    def __init__(self, config, cache_data):
        self._log = logging.getLogger('crosspm')
        self._config = config
        self._cache_path = config.crosspm_cache_root
        if not os.path.exists(self._cache_path):
            os.makedirs(self._cache_path)

        self.path = {
            'packed': os.path.realpath(os.path.join(self._cache_path, 'archive')),
            'unpacked': os.path.realpath(os.path.join(self._cache_path, 'cache')),
        }
        self.temp_path = os.path.realpath(os.path.join(self._cache_path, 'tmp'))

        self._clear = cache_data.get('clear', {})
        self._clear['auto'] = self._clear.get('auto', False)
        if isinstance(self._clear['auto'], str):
            if self._clear['auto'].lower() in ['0', 'no', '-', 'false']:
                self._clear['auto'] = False
        try:
            self._clear['auto'] = bool(self._clear['auto'])
        except Exception:
            self._clear['auto'] = False

        self._clear['size'] = self.str_to_size(self._clear.get('size', '0'))

        try:
            self._clear['days'] = int(self._clear.get('days', '0'))
        except Exception:
            self._clear['days'] = 0

        _unique = config.get_fails('unique', None)
        self._storage = cache_data.get('storage', {'packed': '', 'unpacked': ''})
        if not self._storage['packed']:
            if _unique and isinstance(_unique, (list, tuple)):
                self._storage['packed'] = '/'.join('{%s}' % x for x in _unique)
                if self._storage['packed']:
                    self._storage['packed'] += '/{filename}'

        if not self._storage['packed']:
            self._storage['packed'] = '{%s}/{filename}' % config.name_column

        if not self._storage['unpacked']:
            if _unique and isinstance(_unique, (list, tuple)):
                self._storage['unpacked'] = '/'.join('{%s}' % x for x in _unique)

        if not self._storage['unpacked']:
            self._storage['unpacked'] = '{%s}' % config.name_column

    def str_to_size(self, size):
        _size = str(size).replace(' ', '').lower()
        try:
            _size0 = float(_size[:-2])
        except Exception:
            _size0 = 0
        _measure = _size[-2:]
        if _measure == 'kb':
            _size = _size0 * 1024
        elif _measure == 'mb':
            _size = _size0 * 1024 * 1024
        elif _measure == 'gb':
            _size = _size0 * 1024 * 1024 * 1024
        else:
            if _size[-1] == 'b':
                _size = _size[:-1]
            try:
                _size = float(_size)
            except Exception:
                _size = 0
        return _size

    def size_to_str(self, size, dec=0):
        try:
            _size = float(size)
        except Exception:
            _size = 0
        _measure = 'b '
        _size0 = round(_size / 1024, 3)
        if _size0 >= 1.0:
            _measure = 'Kb'
            _size = _size0
            _size0 = round(_size / 1024, 3)
            if _size0 >= 1.0:
                _measure = 'Mb'
                _size = _size0
                _size0 = round(_size / 1024, 3)
                if _size0 >= 1.0:
                    _measure = 'Gb'
                    _size = _size0
        _size = round(_size, dec)
        if dec == 0:
            _size = int(_size)
        return '{:,} {}'.format(_size, _measure)

    def get_info(self, get_files=True):
        def get_dir(_dir_name, sub_dirs=True):
            res = {
                'path': _dir_name,
                'size': 0,
                'time': 0,
                'files': [],
                'folders': [],
            }
            if os.path.isdir(_dir_name):
                for item in os.listdir(_dir_name):
                    item_name = os.path.realpath(os.path.join(_dir_name, item))
                    item_stat = os.stat(item_name)
                    _size = getattr(item_stat, 'st_size', 0)
                    _now = datetime.now()
                    _time = getattr(item_stat, 'st_ctime',
                                    time.mktime(_now.timetuple()) + float(_now.microsecond) / 1000000.0)
                    if os.path.isfile(os.path.join(_dir_name, item)):
                        res['size'] += _size
                        res['time'] = max(res['time'], _time)
                        res['files'].append({'size': _size,
                                             'time': _time,
                                             'path': item_name})
                    elif sub_dirs:
                        _folder = get_dir(item_name)
                        res['folders'].append(_folder)
                        res['size'] += _folder['size']
                        res['time'] = max(res['time'], _folder['time'])
            return res

        total = {
            'packed': get_dir(self.path['packed']),
            'unpacked': get_dir(self.path['unpacked']),
            'temp': get_dir(self.temp_path),
            'other': get_dir(self._cache_path, False),
        }
        return total

    def _sort(self, item):
        return item['time']

    def info(self):
        print("Cache info:")

    def _delete_dir(self, _dirs, max_time=None, del_size=None):
        cleared = [0]

        def do_delete_dir(_dir):
            _size = 0
            folders_to_delete = []
            files_to_delete = []
            for _folder in _dir['folders']:
                _size += do_delete_dir(_folder)
                if len(_folder['folders']) + len(_folder['files']) == 0:
                    os.rmdir(_folder['path'])
                    folders_to_delete += [_folder]
            for _file in _dir['files']:
                do_clear = False
                if max_time:
                    if max_time > _file['time']:
                        do_clear = True
                if del_size:
                    if del_size > cleared[0]:
                        do_clear = True
                if not (max_time or del_size):
                    do_clear = True
                if do_clear:
                    _size += _file['size']
                    cleared[0] += _file['size']
                    os.remove(_file['path'])
                    files_to_delete += [_file]
            _dir['size'] -= _size
            for _folder in folders_to_delete:
                _dir['folders'].remove(_folder)
            for _file in files_to_delete:
                _dir['files'].remove(_file)
            return _size

        return do_delete_dir(_dirs)

    def auto_clear(self):
        if self._clear['auto']:
            self.clear(False, True)

    def clear(self, hard=False, auto=False):

        # TODO: Check if given path is really crosspm cache before deleting anything!
        _now = datetime.now().timestamp()
        self._log.info('Read cache...')
        max_size = self._clear['size']
        if self._clear['days']:
            max_time = _now - timedelta(days=self._clear['days']).total_seconds()
        else:
            max_time = 0

        total = self.get_info()

        total_size = sum(total[x]['size'] for x in total)
        oldest = min(total[x]['time'] if total[x]['time'] else _now for x in total)

        if auto and not hard:
            do_clear = False
            if max_size and max_size < total_size:
                do_clear = True
            if max_time and max_time > oldest:
                do_clear = True
            if not do_clear:
                return

        self._log.info('Clear cache{}'.format(' HARD! >:-E' if hard else ':'))

        self.size(total)

        # TEMP
        cleared = self._delete_dir(total['temp'])

        # UNPACKED
        if hard:
            cleared += self._delete_dir(total['unpacked'])
        else:
            folders_to_delete = []
            _size = 0
            for folder in sorted(total['unpacked']['folders'], key=self._sort):
                do_clear = False
                if max_size:
                    if total_size - cleared - _size > max_size:
                        do_clear = True
                if max_time:
                    if folder['time'] < max_time:
                        do_clear = True
                if do_clear:
                    _size += self._delete_dir(folder)
                if len(folder['folders']) + len(folder['files']) == 0:
                    os.rmdir(folder['path'])
                    folders_to_delete += [folder]
                    # if (not do_clear) and (not max_time):
                    #     break
            total['unpacked']['size'] -= _size
            cleared += _size
            for folder in folders_to_delete:
                total['unpacked']['folders'].remove(folder)

        # PACKED
        if hard:
            cleared += self._delete_dir(total['packed'])
        else:
            del_size = total_size - cleared - max_size
            if del_size < 0:
                del_size = 0
            if del_size or max_time:
                cleared += self._delete_dir(total['packed'], max_time, del_size)

        self._log.info('')
        self._log.info('Очищено: {}'.format(self.size_to_str(cleared, 1)))
        self._log.info('')
        self.size(total)

    def size(self, total=None):
        if not total:
            total = self.get_info(False)
        self._log.info('Cache size:')
        self._log.info('    packed: {:>15}'.format(self.size_to_str(total['packed']['size'], 1)))
        self._log.info('  unpacked: {:>15}'.format(self.size_to_str(total['unpacked']['size'], 1)))
        self._log.info('      temp: {:>15}'.format(self.size_to_str(total['temp']['size'], 1)))
        self._log.info('     other: {:>15}'.format(self.size_to_str(total['other']['size'], 1)))
        self._log.info('---------------------------')
        self._log.info('     total: {:>15}'.format(self.size_to_str(sum(total[x]['size'] for x in total), 1)))

    def age(self, total=None):
        if not total:
            total = self.get_info(False)
        oldest = min((x for x in (
            total['packed']['time'], total['unpacked']['time'], total['temp']['time'], total['other']['time']) if x))
        self._log.info('Cache oldest timestamp:')
        self._log.info('    packed: {}'.format(datetime.fromtimestamp(total['packed']['time'])))
        self._log.info('  unpacked: {}'.format(datetime.fromtimestamp(total['unpacked']['time'])))
        self._log.info('      temp: {}'.format(datetime.fromtimestamp(total['temp']['time'])))
        self._log.info('     other: {}'.format(datetime.fromtimestamp(total['other']['time'])))
        self._log.info('---------------------------')
        self._log.info('    oldest: {}'.format(datetime.fromtimestamp(oldest)))

    def path_packed(self, package=None, params=None):
        return self.path_any('packed', package, params)

    def path_unpacked(self, package=None, params=None):
        return self.path_any('unpacked', package, params)

    def path_any(self, name, package=None, params=None):
        if params:
            tmp_params = {}
            for k, v in params.items():
                if isinstance(v, list):
                    v = ".".join([x for x in v if x is not None and x != ''])
                tmp_params[k] = v
            res = self._storage[name].format(**tmp_params)
        elif package:
            res = self._storage[name].format(**(package.get_params(merged=True)))
        else:
            res = ''
        res = os.path.realpath(os.path.join(self.path[name], res))
        return res

    def exists_packed(self, package=None, params=None, pkg_path='', check_stat=True):
        # Check if file exists and size and time match
        path = self.path_packed(package, params) if not pkg_path else pkg_path
        res = os.path.exists(path)
        if res and package and check_stat:
            _stat_attr = {'ctime': 'st_atime',
                          'mtime': 'st_mtime',
                          'size': 'st_size'}
            _stat_file = os.stat(path)
            if any(package.stat[k] != getattr(_stat_file, v, -999) for k, v in _stat_attr.items()):
                res = False
                os.remove(path)

        return res, path

    def exists_unpacked(self, package, filename=None, params=None, pkg_path=''):
        # TODO: Check if file exists and size and time match
        path = self.path_unpacked(package, params) if not pkg_path else pkg_path
        if os.path.isdir(path):
            # check that dir is not empty
            res = bool(os.listdir(path))
        else:
            res = os.path.exists(path)

        return res, path



# -*- coding: utf-8 -*-
import os

from crosspm.helpers.config import CROSSPM_DEPENDENCY_FILENAME
from crosspm.helpers.content import DependenciesContent
from crosspm.helpers.downloader import Downloader
from crosspm.helpers.output import Output


class Locker(Downloader):
    def __init__(self, config, do_load, recursive):
        # TODO: revise logic to allow recursive search without downloading
        super(Locker, self).__init__(config, do_load, recursive)

        if not getattr(config, 'deps_path', ''):
            config.deps_path = config.deps_file_name or CROSSPM_DEPENDENCY_FILENAME

        deps_path = config.deps_path
        if deps_path.__class__ is DependenciesContent:
            # HACK
            pass
            self._deps_path = deps_path
        else:
            deps_path = config.deps_path.strip().strip('"').strip("'")
            self._deps_path = os.path.realpath(os.path.expanduser(deps_path))

    def lock_packages(self, deps_file_path=None, depslock_file_path=None, packages=None):
        """
        Lock packages. Downloader search packages
        """
        if deps_file_path is None:
            deps_file_path = self._deps_path
        if depslock_file_path is None:
            depslock_file_path = self._depslock_path
        if deps_file_path == depslock_file_path:
            depslock_file_path += '.lock'
            # raise CrosspmException(
            #     CROSSPM_ERRORCODE_WRONG_ARGS,
            #     'Dependencies and Lock files are same: "{}".'.format(deps_file_path),
            # )

        if packages is None:
            self.search_dependencies(deps_file_path)
        else:
            self._root_package.packages = packages

        self._log.info('Writing lock file [{}]'.format(depslock_file_path))

        output_params = {
            'out_format': 'lock',
            'output': depslock_file_path,
        }
        Output(config=self._config).write_output(output_params, self._root_package.packages)
        self._log.info('Done!')

    def entrypoint(self, *args, **kwargs):
        self.lock_packages(*args, **kwargs)



# -*- coding: utf-8 -*-
import json
import logging
import os
import platform
import sys

import requests
import yaml

from crosspm.helpers.cache import Cache
from crosspm.helpers.content import DependenciesContent
from crosspm.helpers.exceptions import *
from crosspm.helpers.parser import Parser
from crosspm.helpers.source import Source

requests.packages.urllib3.disable_warnings()

WINDOWS = (platform.system().lower() == 'windows') or (os.name == 'nt')
DEFAULT_CONFIG_FILE = ('crosspm.yaml', 'crosspm.yml', 'crosspm.json',)
USER_HOME_DIR = os.path.expanduser('~')
CROSSPM_ROOT_DIR = os.path.abspath(os.path.dirname(os.path.dirname(__file__)))
DEFAULT_CONFIG_PATH = [
    './',
    '~/',
    '~/.crosspm',
    '/etc/crosspm',
    CROSSPM_ROOT_DIR,
] if not WINDOWS else [
    './',
    os.path.realpath(os.path.join(os.path.splitdrive(USER_HOME_DIR)[0], '/')),
    os.path.realpath(os.path.join(USER_HOME_DIR, '.crosspm')),
    os.path.realpath(os.path.join(os.getenv('APPDATA'), 'crosspm')),
    CROSSPM_ROOT_DIR,
]
GLOBAL_CONFIG_FILE = ['crosspm_global.yaml', 'global.yaml']
GLOBAL_CONFIG_PATH = [
    '/etc/crosspm',
] if not WINDOWS else [
    '{system_drive}',
    '{all_users}\\crosspm',
    '{program_data}\\crosspm',
]

ENVIRONMENT_CONFIG_PATH = 'CROSSPM_CONFIG_PATH'
CROSSPM_DEPENDENCY_FILENAME = 'dependencies.txt'  # maybe 'cpm.manifest'
CROSSPM_DEPENDENCY_LOCK_FILENAME = CROSSPM_DEPENDENCY_FILENAME  # 'dependencies.txt.lock'
CROSSPM_ADAPTERS_NAME = 'adapters'
CROSSPM_ADAPTERS_DIR = os.path.join(CROSSPM_ROOT_DIR, CROSSPM_ADAPTERS_NAME)


class Config:
    windows = WINDOWS

    def __init__(self, config_file_name='', cmdline='', no_fails=False, depslock_path='', deps_path='',
                 lock_on_success=False, prefer_local=False):
        self._log = logging.getLogger('crosspm')
        self._config_path_env = []
        self._sources = []
        self._adapters = {}
        self._parsers = {}
        self._defaults = {}
        self._columns = []
        self._not_columns = {}
        self._options = {}
        self._values = {}
        self._output = {}
        self._solid = {}
        self._fails = {}
        self.name_column = ''
        self.deps_file_name = ''
        self.deps_lock_file_name = ''
        self.lock_on_success = lock_on_success
        self.prefer_local = prefer_local
        self.crosspm_cache_root = ''
        self.deps_path = ''
        self.depslock_path = ''
        self.cache_config = {}
        self.init_env_config_path()
        self.secret_variables = []

        cpm_conf_name = ''
        if deps_path:
            if deps_path.__class__ is DependenciesContent:
                # HACK
                self.deps_path = deps_path
            else:
                deps_path = deps_path.strip().strip('"').strip("'")
                self.deps_path = os.path.realpath(os.path.expanduser(deps_path))
            if not cpm_conf_name:
                cpm_conf_name = self.get_cpm_conf_name(deps_path)
            if os.path.isfile(deps_path):
                config_path_tmp = os.path.dirname(deps_path)
            else:
                config_path_tmp = deps_path
            if config_path_tmp not in DEFAULT_CONFIG_PATH:
                DEFAULT_CONFIG_PATH.append(config_path_tmp)

        if depslock_path:
            if depslock_path.__class__ is DependenciesContent:
                # HACK
                self.depslock_path = depslock_path
            else:
                depslock_path = depslock_path.strip().strip('"').strip("'")
                self.depslock_path = os.path.realpath(os.path.expanduser(depslock_path))

            if not cpm_conf_name:
                cpm_conf_name = self.get_cpm_conf_name(depslock_path)
            if os.path.isfile(depslock_path):
                config_path_tmp = os.path.dirname(depslock_path)
            else:
                config_path_tmp = depslock_path
            if config_path_tmp not in DEFAULT_CONFIG_PATH:
                DEFAULT_CONFIG_PATH.append(config_path_tmp)

        self._config_file_name = self.find_config_file(config_file_name, cpm_conf_name)
        self._global_config_file_name = self.find_global_config_file()
        if self._global_config_file_name:
            config_data = self.read_config_file(True)
            _override = config_data.pop('override', True)
            if isinstance(_override, str):
                if _override.lower() in ['0', 'no', '-', 'false']:
                    _override = False
            try:
                _override = bool(_override)
            except Exception:
                _override = True
        else:
            config_data = {}
            _override = False
        if _override:
            config_data.update({k: v for k, v in self.read_config_file().items() if k not in config_data})
            self._log.debug('Overriding config file values with global config.')
        else:
            config_data.update(self.read_config_file())

        # Add secret variables to special list in config
        if 'options' in config_data:
            for variable, property in config_data['options'].items():
                if property.get('secret', False):
                    self.secret_variables.append(variable)

        self.no_fails = no_fails
        self.parse_config(config_data, cmdline)
        self.cache = Cache(self, self.cache_config)
        # self._fails = {}

    def get_cpm_conf_name(self, deps_filename=''):
        if not deps_filename:
            deps_filename = self.depslock_path
        result = ''
        if os.path.isfile(deps_filename):
            try:
                with open(deps_filename, 'r', encoding="utf-8-sig") as f:
                    for line in f:
                        line = line.strip()
                        if line.startswith('#'):
                            line = [x.strip().strip('"').strip("'") for x in line.strip('#').split('=') if x]
                            if len(line) > 1:
                                if line[0].lower() == 'cpmconfig':
                                    result = line[1].split('#')[0].strip('"').strip("'")
                                    break
            except Exception:
                pass
        return result

    def init_env_config_path(self):
        self._config_path_env = [x for x in os.getenv(ENVIRONMENT_CONFIG_PATH, '').split(';') if x]
        for config_path_env in self._config_path_env:
            config_path_tmp = os.path.dirname(config_path_env) if os.path.isfile(config_path_env) else config_path_env
            if config_path_tmp in DEFAULT_CONFIG_PATH:
                DEFAULT_CONFIG_PATH.remove(config_path_tmp)
            DEFAULT_CONFIG_PATH.insert(0, config_path_tmp)

    def find_global_config_file(self):
        args = {}
        _win_disk = self.get_windows_system_disk()
        if WINDOWS:
            args = {
                'system_drive': _win_disk,
                'all_users': os.getenv('AllUsersProfile', os.path.join(_win_disk, 'Users\\All Users')),
                'program_data': os.getenv('ProgramData', os.path.join(_win_disk, 'ProgramData')),
            }

        for config_path_env in self._config_path_env:
            if os.path.isdir(config_path_env):
                if config_path_env in GLOBAL_CONFIG_PATH:
                    GLOBAL_CONFIG_PATH.remove(config_path_env)
                GLOBAL_CONFIG_PATH.insert(0, config_path_env)

        for config_path in GLOBAL_CONFIG_PATH:
            try:
                _path = config_path.format(**args).strip().strip("'").strip('"')
            except Exception:
                _path = ''
            if _path:
                for config_name in GLOBAL_CONFIG_FILE:
                    _file = os.path.realpath(os.path.expanduser(os.path.join(_path, config_name)))
                    if os.path.isfile(_file):
                        return _file
        return ''

    @staticmethod
    def find_cpmconfig(conf_name=''):
        conf_path_add = ''
        try:
            _temp = __import__('cpmconfig.cpmconfig', globals(), locals(), ['App'], 0)
            cpm_find = _temp.App.find
            _temp_conf = __import__('cpmconfig.conf', globals(), locals(), ['conf', 'conf_file', 'conf_path'], 0)
            conf_path = getattr(_temp_conf, 'conf_path', '')
            if not conf_path:
                conf_file = getattr(_temp_conf, 'conf_file', '')
                if not conf_file:
                    conf = getattr(_temp_conf, 'conf', '')
                    if conf:
                        conf_path = cpm_find(None, conf)[2]
                else:
                    conf_path = cpm_find(None, conf_file)[2]
            if conf_name:
                conf_path_add = cpm_find(None, conf_name)[2]

        except Exception:
            conf_path = ''
        if conf_path and not os.path.isfile(conf_path):
            conf_path = ''
        if conf_path_add and not os.path.isfile(conf_path_add):
            conf_path_add = ''
        return conf_path, conf_path_add

    def find_config_file(self, config_file_name='', cpm_conf_name=''):
        if not config_file_name:
            ind = 0
            for config_path_env in self._config_path_env:
                if os.path.isfile(config_path_env):
                    if config_path_env in DEFAULT_CONFIG_PATH:
                        DEFAULT_CONFIG_PATH.remove(config_path_env)
                    DEFAULT_CONFIG_PATH.insert(ind, config_path_env)
                    ind += 1

            cpm_file_default, cpm_file_deps = self.find_cpmconfig(cpm_conf_name)
            if cpm_file_deps:
                DEFAULT_CONFIG_PATH.insert(ind, cpm_file_deps)

            _def_conf_file = [DEFAULT_CONFIG_FILE] if isinstance(DEFAULT_CONFIG_FILE, str) else DEFAULT_CONFIG_FILE
            for config_path in DEFAULT_CONFIG_PATH:
                config_path = config_path.strip().strip("'").strip('"')
                config_path = os.path.realpath(os.path.expanduser(config_path))
                for _conf_file in _def_conf_file:
                    config_file_name = os.path.realpath(os.path.join(config_path, _conf_file)) if os.path.isdir(
                        config_path) else config_path
                    if os.path.isfile(config_file_name):
                        break
                    else:
                        config_file_name = ''
                if config_file_name:
                    break

            if not config_file_name:
                config_file_name = cpm_file_default
            if config_file_name:
                self._log.info('Found config file [%s]', config_file_name)
        else:
            config_file_name = config_file_name.strip().strip("'").strip('"')
            config_file_name = os.path.realpath(os.path.expanduser(config_file_name))
            self._log.info('Check config file [%s]', config_file_name)
            if not os.path.isfile(config_file_name):
                config_file_name = ''
            else:
                config_path_tmp = os.path.dirname(config_file_name)
                if config_path_tmp not in DEFAULT_CONFIG_PATH:
                    DEFAULT_CONFIG_PATH.append(config_path_tmp)

        if config_file_name == '':
            raise CrosspmException(
                CROSSPM_ERRORCODE_CONFIG_NOT_FOUND,
                'Config file does not found',
            )

        return config_file_name

    def read_config_file(self, _global=False):
        if _global:
            _config_file_name = self._global_config_file_name
            self._log.info('Reading global config file...')
        else:
            _config_file_name = self._config_file_name
            self._log.info('Reading config file... [%s]', _config_file_name)
        _ext = os.path.splitext(_config_file_name)[1].lower()
        _is_yaml = True
        if _ext in ['.yaml', '.yml']:
            _is_yaml = True
        elif _ext == '.json':
            _is_yaml = False
        else:
            with open(_config_file_name, encoding="utf-8-sig") as f:
                for line in f:
                    line = line.strip()
                    if not line:
                        continue
                    if line.startswith('{'):
                        _is_yaml = False  # Assuming JSON format
                        break

        try:
            if _is_yaml:
                result = self.load_yaml(_config_file_name)
            else:
                with open(_config_file_name, encoding="utf-8-sig") as f:
                    result = json.loads(f.read())

        except Exception as e:
            self._log.exception(e)
            code = CROSSPM_ERRORCODE_CONFIG_IO_ERROR
            msg = 'Error reading config file: [{}]'.format(self._config_file_name)
            raise CrosspmException(code, msg) from e

        return result

    @staticmethod
    def find_import_file(import_file_name=''):
        res_file_name = ''
        if import_file_name:
            for config_path in [''] + DEFAULT_CONFIG_PATH:
                if config_path:
                    config_path = config_path.strip().strip("'").strip('"')
                    config_path = os.path.realpath(os.path.expanduser(config_path))
                if config_path == '' or os.path.isdir(config_path):
                    res_file_name = os.path.realpath(
                        os.path.expanduser(import_file_name) if config_path == '' else
                        os.path.join(config_path, import_file_name))
                    if os.path.isfile(res_file_name):
                        break
                    else:
                        res_file_name = ''
                else:
                    res_file_name = ''

            if res_file_name == '':
                raise CrosspmException(
                    CROSSPM_ERRORCODE_CONFIG_NOT_FOUND,
                    'Config import file "{}" does not found'.format(import_file_name),
                )

        return res_file_name

    @staticmethod
    def load_yaml(_config_file_name):
        result = {}
        yaml_imports = ''
        yaml_content = ''
        with open(_config_file_name, encoding="utf-8-sig") as f:
            for line in f:
                if line.strip().startswith('#'):
                    continue
                if (not yaml_imports) and (not yaml_content):
                    line_one = line.strip().replace(' ', '').lower()
                    if line_one.startswith('import:'):
                        yaml_imports += line
                    elif len(line_one) > 0:
                        yaml_content += line
                elif yaml_imports and not yaml_content:
                    if line.startswith((' ', '\t',)):
                        yaml_imports += line
                    elif line.strip():
                        yaml_content += line
                else:
                    yaml_content += line
        if yaml_imports:
            data_imports = yaml.safe_load(yaml_imports)
            yaml_content_pre = ''
            if 'import' in data_imports:
                if isinstance(data_imports['import'], str):
                    data_imports['import'] = [data_imports['import']]
                if isinstance(data_imports['import'], (list, tuple)):
                    for _import_file_name in data_imports['import']:
                        _import_file_name = Config.find_import_file(_import_file_name)
                        if _import_file_name:
                            with open(_import_file_name, encoding="utf-8-sig") as f:
                                for line in f:
                                    yaml_content_pre += line
                            yaml_content_pre += '\n'
            yaml_content = yaml_content_pre + yaml_content
        if yaml_content:
            result = yaml.safe_load(yaml_content)
        return result

    def parse_config(self, config_data, cmdline):
        # init parsers
        if 'parsers' in config_data:
            self.init_parsers(config_data['parsers'])

        # init columns
        if 'columns' in config_data:
            self._columns = []
            _name_index = 0
            for i, _col in enumerate([x for x in [x.strip().split(' ') for x in config_data['columns'].split(',')]]):
                if _col[0][0] == '*':
                    _name_index = i
                    _col = [_col[0][1:]]
                self._columns += _col
            self.name_column = self._columns[_name_index]

        # init available values for columns
        if 'values' in config_data:
            self._values = config_data['values']

        # init options
        if 'options' in config_data:
            self._options = config_data['options']

        # init default values for columns
        if 'defaults' in config_data:
            self._defaults = config_data['defaults']
        self._defaults.update(
            {k: v['default'] for k, v in self._options.items() if k not in self._defaults and 'default' in v})

        # init solid
        if 'solid' in config_data:
            self._solid = config_data['solid']
            if 'ext' in self._solid and not isinstance(self._solid['ext'], (list, tuple)):
                self._solid['ext'] = [self._solid['ext']]

        # init fails
        if 'fails' in config_data:
            self._fails = config_data['fails']

        # init output
        if 'output' in config_data:
            self._output = config_data['output']

        # init common parameters
        _common = {}
        if 'common' in config_data:
            _common = config_data['common']

        if 'sources' in config_data:

            # init adapters
            _types = []
            if 'type' in _common:
                _types += [_common['type']]
            for _src in config_data['sources']:
                if 'type' in _src and _src['type'] not in _types:
                    _types += [_src['type']]

            self.init_adapters(_types)

            # init sources
            for _src in config_data['sources']:
                _src.update({k: v for k, v in _common.items() if k not in _src})

                if 'type' in _src:
                    if _src['type'] not in self._adapters.keys():
                        code = CROSSPM_ERRORCODE_CONFIG_FORMAT_ERROR
                        msg = 'Adapter [{}] does not found!'.format(_src['type'])
                        self._log.exception(msg)
                        raise CrosspmException(code, msg)
                else:
                    code = CROSSPM_ERRORCODE_CONFIG_FORMAT_ERROR
                    msg = 'Source must have type property!'
                    self._log.exception(msg)
                    raise CrosspmException(code, msg)

                if 'parser' in _src:
                    if _src['parser'] not in self._parsers.keys():
                        code = CROSSPM_ERRORCODE_CONFIG_FORMAT_ERROR
                        msg = 'Parser [{}] does not found!'.format(_src['parser'])
                        self._log.exception(msg)
                        raise CrosspmException(code, msg)

                self._sources.append(
                    Source(
                        self._adapters[_src['type']],
                        self._parsers[_src['parser']],
                        _src
                    )
                )

        if len(self._sources) == 0:
            code = CROSSPM_ERRORCODE_CONFIG_FORMAT_ERROR
            msg = 'No correct sources defined! Unable to process any further.'
            self._log.exception(msg)
            raise CrosspmException(code, msg)

        try:
            _cmdline = {x[0]: x[1] for x in [x.strip().split('=') for x in cmdline.split(',')] if len(x) > 1}
        except Exception:
            _cmdline = {}

        # init cpm parameters
        for x in ['cpm', 'crosspm']:
            if x in config_data:
                # init cache:
                if 'cache' not in config_data:
                    config_data['cache'] = {}
                self.init_cpm_and_cache(config_data[x], _cmdline, config_data['cache'])
                break

        self._options = self.parse_options(self._options, _cmdline)

        # init not columns:
        self.init_not_columns()

    def init_cpm_and_cache(self, crosspm, cmdline, cache_config):
        if 'cache' not in crosspm:
            crosspm['cache'] = {}

        # have to merge dicts:
        parts = ('clear', 'storage')
        cache_config.update(
            {k: v for k, v in crosspm['cache'].items() if (k not in cache_config) and (k not in parts)})
        for part in parts:
            if part in crosspm['cache']:
                if part not in cache_config:
                    cache_config[part] = {}
                cache_config[part].update(
                    {k: v for k, v in crosspm['cache'][part].items() if k not in cache_config[part]})

        if 'path' in cache_config:
            crosspm['cache'] = cache_config.pop('path')
        else:
            for x in ['env', 'cmdline', 'default']:
                if x in cache_config:
                    crosspm['cache'][x] = cache_config.pop(x)

        crosspm = self.parse_options(crosspm, cmdline)

        self.deps_lock_file_name = crosspm.get('dependencies-lock', CROSSPM_DEPENDENCY_LOCK_FILENAME)
        if 'dependencies' in crosspm:
            self.deps_file_name = crosspm.get('dependencies', CROSSPM_DEPENDENCY_FILENAME)
            if 'dependencies-lock' not in crosspm:
                self.deps_lock_file_name = self.deps_file_name

        if not self.prefer_local:
            self.prefer_local = crosspm.get('prefer-local', False)
        if not self.lock_on_success:
            lock_on_success = crosspm.get('lock-on-success', False)
            if isinstance(lock_on_success, str):
                if lock_on_success.lower() in ['1', 'yes', '+', 'true']:
                    self.lock_on_success = True
            try:
                self.lock_on_success = bool(lock_on_success)
            except Exception:
                self.lock_on_success = False

        # Cache init
        self.crosspm_cache_root = crosspm.get('cache', '').strip().strip('"').strip("'")
        if not self.crosspm_cache_root:
            home_dir = os.getenv('APPDATA') if WINDOWS else os.getenv('HOME')
            self.crosspm_cache_root = os.path.join(home_dir, '.crosspm')
        self.crosspm_cache_root = os.path.realpath(os.path.expanduser(self.crosspm_cache_root))
        self.cache_config = cache_config

    def init_not_columns(self):
        # gather items from options
        # include options here even if they are columns also
        self._not_columns = {k: v for k, v in self._options.items()}

        # gather items from defaults
        self._not_columns.update({k: v for k, v in self._defaults.items() if k not in self._not_columns})

        # gather items from parsers
        for _parser in self._parsers.values():
            self._not_columns.update({k: None for k in _parser.get_vars() if k not in self._not_columns})

    def init_parsers(self, parsers):
        if 'common' not in parsers:
            parsers['common'] = {}
        for k, v in parsers.items():
            if k != 'common':
                if k not in self._parsers:
                    v.update({_k: _v for _k, _v in parsers['common'].items() if _k not in v})
                    self._parsers[k] = Parser(k, v, self)
                else:
                    code = CROSSPM_ERRORCODE_CONFIG_FORMAT_ERROR
                    msg = 'Config file contains multiple definitions of the same parser: [{}]'.format(k)
                    self._log.exception(msg)
                    raise CrosspmException(code, msg)
        if len(self._parsers) == 0:
            code = CROSSPM_ERRORCODE_CONFIG_FORMAT_ERROR
            msg = 'Config file does not contain parsers! Unable to process any further.'
            self._log.exception(msg)
            raise CrosspmException(code, msg)

    def get_parser(self, parser_name):
        return self._parsers[parser_name] if parser_name in self._parsers else None

    def init_adapters(self, types):
        if not os.path.isdir(CROSSPM_ADAPTERS_DIR):
            code = CROSSPM_ERRORCODE_ADAPTER_ERROR
            msg = 'Adapters directory does not found!'
            self._log.exception(msg)
            raise CrosspmException(code, msg)

        _cwd = os.getcwd()
        _base_dir = os.path.dirname(CROSSPM_ROOT_DIR)
        _adapters_app = '.'.join(os.path.split(os.path.relpath(CROSSPM_ADAPTERS_DIR, _base_dir)))
        os.chdir(_base_dir)
        _remove = False
        if _base_dir not in sys.path:
            sys.path.insert(0, _base_dir)
            _remove = True

        for _file_name in os.listdir(CROSSPM_ADAPTERS_DIR):
            if _file_name.startswith(('__', 'common',)):
                continue

            _app_file_name, _ext = os.path.splitext(_file_name)
            if not _ext.startswith('.py'):
                continue
            _app_name = '.'.join((_adapters_app, _app_file_name,))

            try:
                _temp = __import__(_app_name, globals(), locals(), ['setup', 'Adapter'], 0)
                _names = _temp.setup['name']
                if isinstance(_names, str):
                    _names = [_names]
                self._adapters.update({k: _temp.Adapter(self) for k in _names if k in types})

            except Exception as e:
                msg = 'Error initializing adapter {}: [{}]'.format(_file_name, e)
                self._log.error(msg)

        if _remove:
            sys.path.remove(_base_dir)
        os.chdir(_cwd)

    @staticmethod
    def parse_options(options, cmdline, check_default=False):
        # try:
        #     _cmdline = {x[0]: x[1] for x in [x.strip().split('=') for x in cmdline.split(',')] if len(x) > 1}
        # except Exception:
        #     _cmdline = {}
        _remove = []
        for k, v in options.items():
            # if option type is str, leave the value intact
            if not isinstance(options[k], (str, bool)):
                # if option has cmdline name, try to fetch a value from command line by cmdline name
                if 'cmdline' in v:
                    if v['cmdline'] in cmdline:
                        options[k] = cmdline[v['cmdline']]

                # if option hasn't cmdline name, try to fetch a value from command line by option's name
                elif k in cmdline:
                    options[k] = cmdline[k]

                # if option is still not filled, try to get a value from environment by env name
                if not isinstance(options[k], str):
                    success = False
                    if 'env' in v:
                        _env = os.getenv(v['env'])
                        if _env:
                            options[k] = _env
                            success = True
                    # if option is still not filled, try to fill with default value if necessary
                    if not success:
                        if check_default and 'default' in v:
                            options[k] = v['default']
                            success = True
                    # if option is still not filled again, just remove it from dict
                    if not success:
                        _remove.append(k)
        for k in _remove:
            options.pop(k)
        return options

    def sources(self):
        for _src in self._sources:
            yield _src

    def check_column_value(self, i, v, get_list=False):
        if i < len(self._columns):
            k = self._columns[i]
            if k in self._options:
                v = self._options[k]
            if v is None:
                if k in self._defaults:
                    v = self._defaults[k]
        else:
            k, v = '', ''
        return [k, v] if get_list else {k: v}

    def get_column_name(self, i):
        return self._columns[i] if len(self._columns) - 1 >= i else None

    def get_columns(self):
        return self._columns

    def complete_params(self, _vars, update=True):
        result = _vars if update else {}
        result.update({k: v for k, v in self._not_columns.items() if k not in _vars})
        result.update({k: '' for k in self._columns if (k not in _vars) and (k not in result)})
        return result

    def get_values(self, column_name):
        _res = None
        if column_name in self._values:
            _res = self._values[column_name]
        return _res

    def iter_valued_columns(self, column_names):
        for column_name in column_names:
            if column_name in self._values:
                yield column_name

    def iter_valued_columns2(self, column_names):
        for column_name in column_names:
            yield column_name, column_name in self._values

    def output(self, out_type, default):
        if out_type in self._output:
            result = self._output[out_type]
        else:
            result = default
        return result

    def get_fails(self, fail_type, default):
        return self._fails.get(fail_type, default)

    def solid(self, package):
        if 'ext' in self._solid:
            if package.ext(self._solid['ext']):
                return True
        return False

    @staticmethod
    def get_verbosity_level(level=None, text=False):
        levels = (
            ('critical', logging.CRITICAL, False),
            ('error', logging.ERROR, False),
            ('warning', logging.WARNING, False),
            ('info', logging.INFO, True),
            ('debug', logging.DEBUG, False),
        )
        if level is None:
            return ', '.join(x[0] for x in levels)

        if level == 'console':
            level = 'debug'

        default = None
        for x in levels:
            if isinstance(level, str):
                if x[0] == level.lower():
                    return x[1]
            elif x[2]:
                default = x[0] if text else x[1]
                break

        return default

    @staticmethod
    def get_windows_system_disk():
        result = os.getenv('SystemDrive', '')
        if not result:
            try:
                _temp = __import__('win32api', globals(), locals(), ['GetSystemDirectory'], 0)
                result = _temp.GetSystemDirectory()
                if result:
                    result = os.path.splitdrive(result)[0]
            except Exception:
                pass
        if not result:
            result = 'C:'
        return result



def get_object_from_string(object_path):
    """
    Return python object from string
    :param object_path: e.g os.path.join
    :return: python object
    """
    # split like crosspm.template.GUS => crosspm.template, GUS
    try:
        module_name, object_name = object_path.rsplit('.', maxsplit=1)
        module_ = __import__(module_name, globals(), locals(), ['App'], 0)
        variable_ = getattr(module_, object_name)
    except Exception:
        variable_ = None
    return variable_



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os

from setuptools import setup
from crosspm import config

build_number = os.getenv('TRAVIS_BUILD_NUMBER', '')
branch = os.getenv('TRAVIS_BRANCH', '')
travis = any((build_number, branch,))
version = config.__version__.split('.')
develop_status = '4 - Beta'
url = 'http://devopshq.github.io/crosspm'

if travis:
    version = version[0:3]
    if branch == 'master':
        develop_status = '5 - Production/Stable'
        version.append(build_number)
    else:
        version.append('{}{}'.format('dev' if branch == 'develop' else branch, build_number))
else:
    if len(version) < 4:
        version.append('dev0')

version = '.'.join(version)
if travis:
    with open('crosspm/config.py', 'w', encoding="utf-8") as f:
        f.write("__version__ = '{}'".format(version))

with open('README.md', encoding="utf-8") as f:
    long_description = f.read()

setup(
    name='crosspm',
    version=version,
    description='Cross Package Manager',
    license='MIT',
    author='Alexander Kovalev',
    author_email='ak@alkov.pro',
    url=url,
    long_description=long_description,
    download_url='https://github.com/devopshq/crosspm.git',
    entry_points={'console_scripts': ['crosspm=crosspm.__main__:main']},
    classifiers=[
        'Development Status :: {}'.format(develop_status),
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.1',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    keywords=[
        'development',
        'dependency',
        'requirements',
        'manager',
        'versioning',
        'packet',
    ],
    packages=[
        'crosspm',
        'crosspm.helpers',
        'crosspm.adapters',
        'crosspm.template',
    ],
    setup_requires=[
        'wheel==0.34.2',
        'pypandoc==1.5',
    ],
    tests_require=[
        "pytest<=4.6.9; python_version < '3.5'",
        "pytest>=5.2; python_version >= '3.5'",
        "pytest-flask<1.0.0; python_version < '3.5'",
        "pytest-flask>=1.0.0; python_version >= '3.5'",
        "PyYAML<5.2; python_version < '3.5'",
        "PyYAML>=5.2; python_version >= '3.5'",
    ],
    install_requires=[
        "requests<2.22; python_version < '3.5'",
        "requests>=2.22; python_version >= '3.5'",
        'urllib3==1.24.3',
        'docopt==0.6.2',
        "PyYAML==5.1.2; python_version < '3.5'",
        "PyYAML>=5.2; python_version >= '3.5'",
        'dohq-artifactory==0.7.377',
        "dohq-artifactory==0.4.112; python_version < '3.5'",
        "dohq-artifactory>=0.7.377; python_version >= '3.5'",
        "Jinja2<2.11; python_version < '3.5'",
        "Jinja2>=2.11; python_version >= '3.5'",
        'patool==1.12',  # need for pyunpack
        'pyunpack==0.1.2',
        # 'pyopenssl>=16.2.0',
        # 'cryptography>=1.7',
    ],
    package_data={
        '': [
            './template/*.j2',
            '*.cmake',
            '../LICENSE',
            # '../README.*',
        ],
    },
)



# -*- coding: utf-8 -*-
import copy
import json
import os
import time
from collections import OrderedDict
from datetime import datetime

import requests
from artifactory import ArtifactoryPath
from requests.auth import HTTPBasicAuth

from crosspm.adapters.common import BaseAdapter
from crosspm.helpers.exceptions import *  # noqa
from crosspm.helpers.package import Package

CHUNK_SIZE = 1024

setup = {
    "name": [
        "artifactory-aql",
        "jfrog-artifactory-aql",
        "artifactory",
        "jfrog-artifactory",
    ],
}

session = requests.Session()


class Adapter(BaseAdapter):
    def get_packages(self, source, parser, downloader, list_or_file_path, property_validate=True):
        """

        :param source:
        :param parser:
        :param downloader:
        :param list_or_file_path:
        :param property_validate: for `root` packages we need check property, bad if we find packages from `lock` file,
        we can skip validate part
        :return:
        """
        _auth_type = source.args['auth_type'].lower() if 'auth_type' in source.args else 'simple'
        _art_auth_etc = {}
        if 'auth' in source.args:
            self.search_auth(list_or_file_path, source)
            if _auth_type == 'simple':
                _art_auth_etc['auth'] = HTTPBasicAuth(*tuple(source.args['auth']))
                session.auth = _art_auth_etc['auth']
                # elif _auth_type == 'cert':
                #     _art_auth_etc['cert'] = os.path.realpath(os.path.expanduser(source.args['auth']))
        if 'auth' not in _art_auth_etc:
            msg = 'You have to set auth parameter for sources with artifactory-aql adapter'
            # self._log.error(msg)
            raise CrosspmException(
                CROSSPM_ERRORCODE_ADAPTER_ERROR,
                msg
            )

        if 'verify' in source.args:
            _art_auth_etc['verify'] = source.args['verify'].lower in ['true', 'yes', '1']
        else:
            _art_auth_etc['verify'] = False

        _pkg_name_column = self._config.name_column
        _secret_variables = self._config.secret_variables
        _packages_found = OrderedDict()
        _pkg_name_old = ""
        _packed_exist = False
        _packed_cache_params = None
        self._log.info('parser: {}'.format(parser._name))
        for _paths in parser.get_paths(list_or_file_path, source):

            # If "parser"-column specified - find only in this parser
            parser_names = _paths['params'].get('parser')
            if parser_names and parser_names != "*":
                self._log.info("Specified parsers: {}".format(parser_names))
                parsers = parser_names.split(',')
                if parser._name not in parsers:
                    self._log.info("Skip parser: {}".format(parser._name))
                    continue

            _packages = []
            _params_found = {}
            _params_found_raw = {}
            last_error = ''
            _pkg_name = _paths['params'][_pkg_name_column]
            if _pkg_name != _pkg_name_old:
                _pkg_name_old = _pkg_name
                self._log.info(
                    '{}: {}'.format(_pkg_name,
                                    {k: v for k, v in _paths['params'].items() if
                                     (k not in (_pkg_name_column, 'repo') and k not in _secret_variables)
                                     }
                                    )
                )
            for _sub_paths in _paths['paths']:
                _tmp_params = dict(_paths['params'])
                self._log.info('repo: {}'.format(_sub_paths['repo']))
                for _path in _sub_paths['paths']:
                    _tmp_params['repo'] = _sub_paths['repo']
                    # ------ START ----
                    # HACK for prefer-local
                    if self._config.prefer_local and not parser.has_rule('properties'):
                        params = parser.get_params_with_extra('path', _paths['params'])
                        for param in params:
                            param['repo'] = _tmp_params['repo']
                            _path_packed = downloader.cache.path_packed(None, param)
                            _packed_exist = os.path.isfile(_path_packed)
                            if _packed_exist:
                                self._log.info("Skip searching, use package cache in path {}".format(_path_packed))
                                _packed_cache_params = param
                                break  # break check local cache
                    if _packed_exist:
                        break  # break connect to artifactory
                        # ------ END  ----
                    _path_fixed, _path_pattern, _file_name_pattern = parser.split_fixed_pattern_with_file_name(_path)
                    try:
                        _artifactory_server = _tmp_params['server']
                        _search_repo = _tmp_params['repo']

                        # Get AQL path pattern, with fixed part path, without artifactory url and repository name
                        _aql_path_pattern = _path_fixed[len(_artifactory_server) + 1 + len(_search_repo) + 1:]
                        if _path_pattern:
                            _aql_path_pattern = _aql_path_pattern + "/" + _path_pattern

                        _aql_query_url = '{}/api/search/aql'.format(_artifactory_server)
                        _aql_query_dict = {
                            "repo": {
                                "$eq": _search_repo,
                            },
                            "path": {
                                "$match": _aql_path_pattern,
                            },
                            "name": {
                                "$match": _file_name_pattern,
                            },
                        }
                        # Remove path if is empty string
                        if not _aql_path_pattern:
                            _aql_query_dict.pop('path')
                        query = 'items.find({query_dict}).include("*", "property")'.format(
                            query_dict=json.dumps(_aql_query_dict))
                        session.auth = _art_auth_etc['auth']
                        r = session.post(_aql_query_url, data=query, verify=_art_auth_etc['verify'])
                        r.raise_for_status()

                        _found_paths = r.json()
                        for _found in _found_paths['results']:
                            _repo_path = "{artifactory}/{repo}/{path}/{file_name}".format(
                                artifactory=_artifactory_server,
                                repo=_found['repo'],
                                path=_found['path'],
                                file_name=_found['name'])
                            _repo_path = ArtifactoryPath(_repo_path, **_art_auth_etc)

                            _mark = 'found'
                            _matched, _params, _params_raw = parser.validate_path(str(_repo_path), _tmp_params)
                            if _matched:
                                _params_found[_repo_path] = {k: v for k, v in _params.items()}
                                _params_found_raw[_repo_path] = {k: v for k, v in _params_raw.items()}
                                _mark = 'match'

                                # Check if it's `root` packages or from `lock` file
                                # ALSO, if from `lock` and have * in name - validate with property
                                property_validate_tmp = property_validate or '*' in _file_name_pattern
                                property_found = 'properties' in _found.keys()
                                # If have not rule in config, skip this part
                                if parser.has_rule('properties') and property_validate_tmp and property_found:
                                    _found_properties = {x['key']: x.get('value', '') for x in _found['properties']}
                                    _valid, _params = parser.validate(_found_properties, 'properties', _tmp_params,
                                                                      return_params=True)
                                else:
                                    _valid, _params = True, {}
                                if _valid:
                                    _mark = 'valid'
                                    _packages += [_repo_path]
                                    _params_found[_repo_path].update({k: v for k, v in _params.items()})
                                    _params_found[_repo_path]['filename'] = str(_repo_path.name)
                                    _params_found[_repo_path]['parser'] = parser._name
                            self._log.debug('  {}: {}'.format(_mark, str(_repo_path)))
                    except RuntimeError as e:
                        try:
                            err = json.loads(e.args[0])
                        except Exception:
                            err = {}
                        if isinstance(err, dict):
                            # Check errors
                            # :e.args[0]: {
                            #                 "errors" : [ {
                            #                     "status" : 404,
                            #                     "message" : "Not Found"
                            #                  } ]
                            #             }
                            for error in err.get('errors', []):
                                err_status = error.get('status', -1)
                                err_msg = error.get('message', '')
                                if err_status == 401:
                                    msg = 'Authentication error[{}]{}'.format(err_status,
                                                                              (': {}'.format(
                                                                                  err_msg)) if err_msg else '')
                                elif err_status == 404:
                                    msg = last_error
                                else:
                                    msg = 'Error[{}]{}'.format(err_status,
                                                               (': {}'.format(err_msg)) if err_msg else '')
                                if last_error != msg:
                                    self._log.error(msg)
                                last_error = msg

            _package = None

            # HACK for prefer-local
            if _packed_exist:
                # HACK - Normalize params for cached archive
                for key, value in _packed_cache_params.items():
                    if isinstance(value, list):
                        value = ['' if x is None else x for x in value]
                        _packed_cache_params[key] = value
                _package = Package(_pkg_name, None, _paths['params'], downloader, self, parser,
                                   _packed_cache_params, list_or_file_path['raw'], {}, in_cache=True)
            # END HACK
            if _packages:
                _tmp = copy.deepcopy(_params_found)
                _packages = parser.filter_one(_packages, _paths['params'], _tmp)
                if isinstance(_packages, dict):
                    _packages = [_packages]

                if len(_packages) == 1:
                    _stat_pkg = self.pkg_stat(_packages[0]['path'])

                    _params_raw = _params_found_raw.get(_packages[0]['path'], {})
                    _params_tmp = _params_found.get(_packages[0]['path'], {})
                    _params_tmp.update({k: v for k, v in _packages[0]['params'].items() if k not in _params_tmp})
                    _package = Package(_pkg_name, _packages[0]['path'], _paths['params'], downloader, self, parser,
                                       _params_tmp, _params_raw, _stat_pkg)
                    _mark = 'chosen'
                    self._log.info('  {}: {}'.format(_mark, str(_packages[0]['path'])))

                elif len(_packages) > 1:
                    raise CrosspmException(
                        CROSSPM_ERRORCODE_MULTIPLE_DEPS,
                        'Multiple instances found for package [{}] not found.'.format(_pkg_name)
                    )
                else:
                    # Package not found: may be error, but it could be in other source.
                    pass
            else:
                # Package not found: may be error, but it could be in other source.
                pass

            if (_package is not None) or (not self._config.no_fails):
                _added, _package = downloader.add_package(_pkg_name, _package)
            else:
                _added = False
            if _package is not None:
                _pkg_name = _package.name
            if _added or (_package is not None):
                if (_package is not None) or (not self._config.no_fails):
                    if (_package is not None) or (_packages_found.get(_pkg_name, None) is None):
                        _packages_found[_pkg_name] = _package

            if _added and (_package is not None):
                if downloader.do_load:
                    _package.download()
                    _deps_file = _package.get_file(self._config.deps_lock_file_name)
                    if downloader.recursive:
                        if _deps_file:
                            _package.find_dependencies(_deps_file, property_validate=False)
                        elif self._config.deps_file_name:
                            _deps_file = _package.get_file(self._config.deps_file_name)
                            if _deps_file and os.path.isfile(_deps_file):
                                _package.find_dependencies(_deps_file, property_validate=False)

        # HACK for not found packages
        _package_names = [x[self._config.name_column] for x in list_or_file_path['raw']]
        _packages_found_names = [x.name for x in _packages_found.values()]
        for package in _package_names:
            if package not in _packages_found_names:
                _packages_found[package] = None

        return _packages_found

    def search_auth(self, list_or_file_path, source):
        """
        Looking for auth in env, cmdline, str
        :param list_or_file_path:
        :param source:
        """
        _auth = source.args['auth']
        if isinstance(_auth, str):
            if ':' in _auth:
                _auth = _auth.split(':')
            elif _auth.endswith('}') and (
                    _auth.startswith('{') or ':' in _auth):  # {auth}, {user}:{password}, user:{password}
                _auth = self.get_auth(list_or_file_path, _auth)
                _auth = self.split_auth(_auth)

        if isinstance(_auth, list):
            for i in range(len(_auth)):
                if _auth[i].endswith('}') and (
                        _auth[i].startswith('{') or ':' in _auth[i]):  # {auth}, {user}:{password}, user:{password}
                    _auth[i] = self.get_auth(list_or_file_path, _auth[i])
                    if ':' in _auth[i]:
                        _auth = self.split_auth(_auth[i])

        source.args['auth'] = _auth

    def get_auth(self, list_or_file_path, _auth):
        try:
            return list_or_file_path['raw'][0][_auth[1:-1]]
        except Exception:
            msg = 'Cred {_auth} not found in options'.format(**locals())
            raise CrosspmException(CROSSPM_ERRORCODE_ADAPTER_ERROR, msg)

    def split_auth(self, _auth):
        if ':' in _auth:
            return _auth.split(':')
        else:
            msg = 'Wrong format of oneline credentials. Use user:password'
            raise CrosspmException(CROSSPM_ERRORCODE_ADAPTER_ERROR, msg)

    @staticmethod
    def pkg_stat(package):
        _stat_attr = {'ctime': 'st_atime',
                      'mtime': 'st_mtime',
                      'size': 'st_size'}
        _stat_pkg = package.stat()
        _stat_pkg = {k: getattr(_stat_pkg, k, None) for k in _stat_attr.keys()}
        _stat_pkg = {
            k: time.mktime(v.timetuple()) + float(v.microsecond) / 1000000.0 if isinstance(v, datetime) else v
            for k, v in _stat_pkg.items()
        }
        return _stat_pkg

    def get_usedby(self, source, parser, downloader, list_or_file_path, property_validate=True):
        """

        :param source:
        :param parser:
        :param downloader:
        :param list_or_file_path:
        :param property_validate: for `root` packages we need check property, bad if we find packages from `lock` file,
        we can skip validate part
        :return:
        """
        _auth_type = source.args['auth_type'].lower() if 'auth_type' in source.args else 'simple'
        _art_auth_etc = {}
        if 'auth' in source.args:
            self.search_auth(list_or_file_path, source)
            if _auth_type == 'simple':
                _art_auth_etc['auth'] = HTTPBasicAuth(*tuple(source.args['auth']))
                session.auth = _art_auth_etc['auth']
                # elif _auth_type == 'cert':
                #     _art_auth_etc['cert'] = os.path.realpath(os.path.expanduser(source.args['auth']))
        if 'auth' not in _art_auth_etc:
            msg = 'You have to set auth parameter for sources with artifactory-aql adapter'
            # self._log.error(msg)
            raise CrosspmException(
                CROSSPM_ERRORCODE_ADAPTER_ERROR,
                msg
            )

        if 'verify' in source.args:
            _art_auth_etc['verify'] = source.args['verify'].lower in ['true', 'yes', '1']
        else:
            _art_auth_etc['verify'] = False

        _secret_variables = self._config.secret_variables
        _pkg_name_col = self._config.name_column
        _packages_found = OrderedDict()
        _pkg_name_old = ""

        for _paths in parser.get_paths(list_or_file_path, source):
            _packages = []
            _params_found = {}
            _params_found_raw = {}
            last_error = ''
            _pkg_name = _paths['params'][_pkg_name_col]
            if _pkg_name != _pkg_name_old:
                _pkg_name_old = _pkg_name
                self._log.info(
                    '{}: {}'.format(_pkg_name,
                                    {k: v for k, v in _paths['params'].items() if
                                     (k not in (_pkg_name_col, 'repo') and k not in _secret_variables)
                                     }
                                    )
                )
            for _sub_paths in _paths['paths']:
                _tmp_params = dict(_paths['params'])
                self._log.info('repo: {}'.format(_sub_paths['repo']))
                _tmp_params['repo'] = _sub_paths['repo']
                try:
                    _artifactory_server = _tmp_params['server']
                    _search_repo = _tmp_params['repo']

                    # TODO: Попробовать использовать lru_cache для кеширования кучи запросов
                    _aql_query_url = '{}/api/search/aql'.format(_artifactory_server)
                    _aql_query_dict = {
                        "repo": {
                            "$eq": _search_repo,
                        },
                    }
                    _usedby_aql = parser.get_usedby_aql(_tmp_params)
                    if _usedby_aql is None:
                        continue
                    _aql_query_dict.update(_usedby_aql)
                    query = 'items.find({query_dict}).include("*", "property")'.format(
                        query_dict=json.dumps(_aql_query_dict))
                    session.auth = _art_auth_etc['auth']
                    r = session.post(_aql_query_url, data=query, verify=_art_auth_etc['verify'])
                    r.raise_for_status()

                    _found_paths = r.json()
                    for _found in _found_paths['results']:
                        _repo_path = "{artifactory}/{repo}/{path}/{file_name}".format(
                            artifactory=_artifactory_server,
                            repo=_found['repo'],
                            path=_found['path'],
                            file_name=_found['name'])
                        _repo_path = ArtifactoryPath(_repo_path, **_art_auth_etc)
                        _found_properties = {x['key']: x.get('value', '') for x in _found['properties']}

                        _matched, _params, _params_raw = parser.validate_path(str(_repo_path), _tmp_params)
                        _params_found[_repo_path] = {k: v for k, v in _params.items()}
                        _params_found_raw[_repo_path] = {k: v for k, v in _params_raw.items()}
                        _params = _tmp_params
                        _packages += [_repo_path]
                        _params_found[_repo_path].update({k: v for k, v in _params.items()})
                        _params_found[_repo_path]['filename'] = str(_repo_path.name)

                        _params_raw = _params_found_raw.get(_repo_path, {})
                        params_found = {}

                        # TODO: Проставление params брать из config.yaml usedby
                        params = parser.get_params_from_properties(_found_properties)
                        params.update(parser.get_params_from_path(str(_repo_path)))
                        _package = Package(params[_pkg_name_col], _repo_path, params, downloader, self,
                                           parser,
                                           params_found, _params_raw)

                        _package.find_usedby(None, property_validate=False)
                        _packages_found[str(_repo_path)] = _package
                        # _package.find_dependencies(_deps_file, property_validate=False)
                        _mark = 'chosen'
                        self._log.info('  {}: {}'.format(_mark, str(_repo_path)))
                except RuntimeError as e:
                    try:
                        err = json.loads(e.args[0])
                    except Exception:
                        err = {}
                    if isinstance(err, dict):
                        # Check errors
                        # :e.args[0]: {
                        #                 "errors" : [ {
                        #                     "status" : 404,
                        #                     "message" : "Not Found"
                        #                  } ]
                        #             }
                        for error in err.get('errors', []):
                            err_status = error.get('status', -1)
                            err_msg = error.get('message', '')
                            if err_status == 401:
                                msg = 'Authentication error[{}]{}'.format(err_status,
                                                                          (': {}'.format(
                                                                              err_msg)) if err_msg else '')
                            elif err_status == 404:
                                msg = last_error
                            else:
                                msg = 'Error[{}]{}'.format(err_status,
                                                           (': {}'.format(err_msg)) if err_msg else '')
                            if last_error != msg:
                                self._log.error(msg)
                            last_error = msg

        return _packages_found

    def download_package(self, package, dest_path):
        self.prepare_dirs(dest_path)

        try:
            _stat_pkg = self.pkg_stat(package)
            # with package.open() as _src:
            session.auth = package.auth
            _src = session.get(str(package), verify=package.verify, stream=True)
            _src.raise_for_status()

            with open(dest_path, 'wb+') as _dest:
                for chunk in _src.iter_content(CHUNK_SIZE):
                    if chunk:  # filter out keep-alive new chunks
                        _dest.write(chunk)
                        _dest.flush()

            _src.close()
            os.utime(dest_path, (_stat_pkg['ctime'], _stat_pkg['mtime']))

        except Exception as e:
            code = CROSSPM_ERRORCODE_SERVER_CONNECT_ERROR
            msg = 'FAILED to download package {} at url: [{}]'.format(
                package.name,
                str(package),
            )
            raise CrosspmException(code, msg) from e

        return dest_path

    def prepare_dirs(self, dest_path):
        dest_dir = os.path.dirname(dest_path)
        if not os.path.exists(dest_dir):
            os.makedirs(dest_dir)
        elif os.path.exists(dest_path):
            os.remove(dest_path)

    @staticmethod
    def get_package_filename(package):
        if isinstance(package, ArtifactoryPath):
            return package.name
        return ''

    @staticmethod
    def get_package_path(package):
        if isinstance(package, ArtifactoryPath):
            return str(package)
        return ''



# -*- coding: utf-8 -*-

import json
import logging
import os
import re
import sys

from jinja2 import Environment, FileSystemLoader
from collections import OrderedDict

from crosspm.helpers.exceptions import *
from crosspm.helpers.parser import Parser

_output_format_map = {}  # Contain map for output format, load from decorator

(
    PLAIN,
    DICT,
    LIST
) = range(3)


class OutFormat:
    def __init__(self, value, esc_path=False):
        self._value = value
        self._esc_path = esc_path

    def __format__(self, fmt):
        result = self._value
        fmts = fmt.split('.')
        if self._esc_path and ('path' not in fmts):
            fmts.insert(0, 'path')
        for fmt in fmts:
            if fmt == 'upper':
                result = str(result).upper()
            elif fmt == 'lower':
                result = str(result).lower()
            elif fmt == 'quote':
                result = '"{}"'.format(result)
            elif fmt == 'unquote':
                result = str(result).strip('"')
            elif fmt == 'safe':
                result = str(result).replace('-', '_')
                result = re.sub(r'[^A-Z_a-z_0-9]', '', result)
                if result:
                    if not (result[0].isalpha() or '_' == result[0]):
                        result = '_' + result
            elif fmt == 'path':
                result = str(result).replace('\\\\', '\\').replace('\\', '\\\\').replace('"', '\\"').replace("'", "\\'")

        return str(result)


def register_output_format(name):
    """
    Load output format function to dictionary (decorator with this function name)
    """

    def check_decorator(fn):
        _output_format_map[name] = fn

        def wrapper(*args, **kwargs):
            return fn(*args, **kwargs)

        return wrapper

    return check_decorator


class Output:
    _output_config = {
        'root': {'PACKAGES_ROOT'},
        # 'key': 'package',
        'value': 'path',
        # 'columns': [
        #     {
        #         'column': 'package',
        #         'value': '{:upper}_ROOT',
        #     },
        #     {
        #         'column': 'path',
        #         'value': '{}',
        #     }
        # ]
    }
    _name_column = ''
    _columns = []

    def __init__(self, data=None, name_column='', config=None):
        self._log = logging.getLogger('crosspm')
        self._config = config
        if name_column:
            self._name_column = name_column
            self._output_config['key'] = name_column
        if data and isinstance(data, dict):
            self._output_config = data
        # self.init_config()
        if 'columns' not in self._output_config:
            self._output_config['columns'] = []

        if len(self._output_config['columns']) == 0:
            # TODO: implement this:
            # if config.no_fails:
            #     param_list = [x for x in config.get_fails('unique', {})]
            #     if self._name_column not in param_list:
            #         param_list.insert(0, self._name_column)
            #     pkg_name = '/'.join('{%s:upper}' % x for x in param_list)
            # else:
            pkg_name = '{:upper}_ROOT'

            self._output_config['columns'] = [
                {
                    'column': self._name_column,
                    'value': pkg_name,
                },
                {
                    'column': 'path',
                    'value': '{}',
                },
            ]

        self.init_config()
        if 'columns' not in self._output_config:
            self._output_config['columns'] = []

    def init_config(self):
        root = self._output_config.get('root', '')
        if isinstance(root, str):
            self._output_config['type'] = PLAIN
        elif isinstance(root, (dict, set)):
            self._output_config['type'] = DICT
            root = [x for x in root]
            self._output_config['root'] = root[0] if len(root) > 0 else ''
        elif isinstance(root, (list, tuple)):
            self._output_config['type'] = LIST
            self._output_config['root'] = root[0] if len(root) > 0 else ''

        key = self._output_config.get('key', '')
        key0 = None

        if 'columns' in self._output_config:
            self._columns = []
            for item in self._output_config['columns']:
                if not item.get('value', ''):
                    item['value'] = '{}'
                if not item.get('name', ''):
                    item['name'] = '{}'
                if not item.get('column', ''):
                    item['column'] = ''
                    for cl in [y for y in [x[0] for x in Parser.split_with_regexp('{.*?}', item['name']) if x[1]] if y]:
                        col = cl.split(':')[0]
                        if col:
                            item['column'] = col
                            break
                if item['column']:
                    self._columns.append(item['column'])
                    if key == item['column']:
                        key0 = ''
                    elif key0 is None:
                        key0 = item['column']

        if key0:
            key = key0
        self._output_config['key'] = key

        if self._columns:
            if self._name_column and self._name_column not in self._columns:
                self._columns.append(self._name_column)
        else:
            if not self._name_column:
                self._name_column = 'package'
        if 'value' not in self._output_config:
            self._output_config['value'] = ''

    @staticmethod
    def get_var_name(pkg_name):
        result = '{:upper.safe}'.format(OutFormat(pkg_name))
        return result

    def write_to_file(self, text, out_file_path):
        out_dir_path = os.path.dirname(out_file_path)

        if not os.path.exists(out_dir_path):
            self._log.info('mkdirs [%s] ...', out_dir_path)
            os.makedirs(out_dir_path)

        elif not os.path.isdir(out_dir_path):
            raise CrosspmException(
                CROSSPM_ERRORCODE_FILE_IO,
                'Unable to make directory [{}]. File with the same name exists'.format(
                    out_dir_path
                ))

        with open(out_file_path, 'w+') as f:
            f.write(text)

    @staticmethod
    def get_output_types():
        return list(_output_format_map.keys())

    def write_output(self, params, packages):
        """
        Функция вызывает определенную функцию для фиксированного out-format
        :param params:
        :param packages:
        :return:
        """
        if params['out_format'] not in _output_format_map:
            raise CrosspmException(
                CROSSPM_ERRORCODE_UNKNOWN_OUT_TYPE,
                'Unknown out_format: [{}]'.format(params['out_format']),
            )

        f = _output_format_map[params['out_format']]
        result = f(self, packages, **params)

        if result:
            out_file_path = os.path.realpath(os.path.expanduser(params['output']))
            self.write_to_file(result, out_file_path)
            self._log.info(
                'Write packages info to file [%s]\ncontent:\n\n%s',
                out_file_path,
                result,
            )

    def format_column(self, column, name, value):
        for item in self._output_config['columns']:
            if item['column'] == column:
                name = item['name'].format(OutFormat(name))
                value = item['value'].format(OutFormat(value))
                break
        return {'name': name, 'value': value}

    @register_output_format('stdout')
    def output_format_stdout(self, packages, **kwargs):
        self._output_config['type'] = PLAIN
        _packages = self.output_format_module(packages)
        for k in _packages:
            v = _packages[k]
            sys.stdout.write('{}: {}\n'.format(self.get_var_name(k), v))
            sys.stdout.flush()
        return None

    @register_output_format('shell')
    def output_format_shell(self, packages, **kwargs):
        self._output_config['type'] = PLAIN
        result = '\n'
        _packages = self.output_format_module(packages)
        for k in _packages:
            v = _packages[k]
            result += "{}='{}'\n".format(self.get_var_name(k), v)
        result += '\n'
        return result

    @register_output_format('jinja')
    def output_format_jinja(self, packages, output_template, **kwargs):
        output_template = os.path.realpath(output_template)
        template_dir, template_name = os.path.split(output_template)
        j2_env = Environment(loader=FileSystemLoader(template_dir), trim_blocks=True)
        result = j2_env.get_template(template_name).render(packages=packages)
        return result

    @register_output_format('cmd')
    def output_format_cmd(self, packages, **kwargs):
        self._output_config['type'] = PLAIN
        result = '\n'
        _packages = self.output_format_module(packages)
        for k in _packages:
            v = _packages[k]
            result += "set {}={}\n".format(self.get_var_name(k), v)
        result += '\n'
        return result

    @register_output_format('lock')
    def output_format_lock(self, packages, **kwargs):
        """ Text to lock file """
        self._output_config['type'] = PLAIN
        text = ''
        tmp_packages = OrderedDict()
        columns = self._config.get_columns()
        widths = {}
        for _pkg in packages.values():
            _pkg_name = _pkg.package_name
            _params = _pkg.get_params(columns, merged=True, raw=False)
            if _pkg_name not in tmp_packages:
                tmp_packages[_pkg_name] = _params
                comment = 1
                for _col in columns:
                    widths[_col] = max(widths.get(_col, len(_col)), len(str(_params.get(_col, '')))) + comment
                    comment = 0
        comment = 1
        for _col in columns:
            text += '{}{} '.format(_col, ' ' * (widths[_col] - len(_col) - comment))
            comment = 0
        text = '#{}\n'.format(text.strip())
        for _pkg_name in sorted(tmp_packages, key=lambda x: str(x).lower()):
            _pkg = tmp_packages[_pkg_name]
            line = ''
            for _col in columns:
                line += '{}{} '.format(_pkg[_col], ' ' * (widths[_col] - len(str(_pkg[_col]))))
            text += '{}\n'.format(line.strip())
        return text

    def output_format_module(self, packages, esc_path=False):
        """
        Create out with child first position
        """

        def create_ordered_list(packages_):
            """
            Recursive for package.packages
            """
            list_ = []
            for _pkg_name in packages_:
                _pkg = packages_[_pkg_name]
                if _pkg and _pkg.packages:
                    list_.extend(create_ordered_list(_pkg.packages))
                if _pkg:
                    _pkg_params = _pkg.get_params(self._columns, True)
                    _res_item = {}
                    for item in self._output_config['columns']:
                        name = item['name'].format(OutFormat(item['column']))
                        value = _pkg_params.get(item['column'], '')
                        if not isinstance(value, (list, dict, tuple)):
                            try:
                                value = item['value'].format(
                                    OutFormat(value, (item['column'] == 'path') if esc_path else False))
                            except Exception:
                                value = ''
                        # TODO: implement this:
                        # if not value:
                        #     try:
                        #         value = item['value'].format(OutFormat(_pkg.get_params('', True)))
                        #     except Exception as e:
                        #         pass
                        _res_item[name] = value
                    list_.append(_res_item)
            return list_

        result_list = create_ordered_list(packages, )

        if self._output_config['type'] == LIST:
            return result_list

        result = OrderedDict()
        for item in result_list:
            # TODO: Error handling
            name = item[self._output_config['key']]
            if self._output_config['value']:
                value = item[self._output_config['value']]
            else:
                value = OrderedDict([(k, v) for k, v in item.items() if k != self._output_config['key']])

            result[name] = value

        return result

    @register_output_format('python')
    def output_format_python(self, packages, **kwargs):
        def get_value(_v):
            if isinstance(_v, (int, float, bool)):
                _res = '{}'.format(str(_v))
            elif isinstance(_v, (dict, tuple, list)):
                _res = '{}'.format(str(_v))
            else:
                _res = "'{}'".format(str(_v))
            return _res

        def items_iter(_dict_or_list):
            if isinstance(_dict_or_list, dict):
                for k in sorted(_dict_or_list, key=lambda x: str(x).lower()):
                    yield k
            else:
                for k in _dict_or_list:
                    yield k

        result = '# -*- coding: utf-8 -*-\n\n'
        res = ''
        dict_or_list = self.output_format_module(packages, True)
        for item in items_iter(dict_or_list):
            if self._output_config['type'] == LIST:
                res += '    {\n'
                for k, v in item.items():
                    res += "        '{}': {},\n".format(k, get_value(v))
                res += '    },\n'
            elif self._output_config['type'] == DICT:
                res += "    '{}': ".format(item)
                if isinstance(dict_or_list[item], dict):
                    res += '{\n'
                    for k, v in dict_or_list[item].items():
                        res += "        '{}': {},\n".format(k, get_value(v))
                    res += '    },\n'
                elif isinstance(dict_or_list[item], (list, tuple)):
                    res += '[\n'
                    for v in dict_or_list[item]:
                        res += "        {},\n".format(get_value(v))
                    res += '    ],\n'
                else:  # str
                    res += "{},\n".format(get_value(dict_or_list[item]))
            else:
                res += '{} = '.format(self.get_var_name(item), get_value(dict_or_list[item]))
                if isinstance(dict_or_list[item], dict):
                    res += '{\n'
                    for k, v in dict_or_list[item].items():
                        res += "    '{}': {},\n".format(k, get_value(v))
                    res += '}\n'
                elif isinstance(dict_or_list[item], (list, tuple)):
                    res += '[\n'
                    for v in dict_or_list[item]:
                        res += "    {},\n".format(get_value(v))
                    res += ']\n'
                else:  # str
                    res += "{}\n".format(get_value(dict_or_list[item]))

        if self._output_config['type'] == LIST:
            result += '{} = [\n'.format(self._output_config['root'] or 'PACKAGES_ROOT')
            result += res
            result += ']\n'
        elif self._output_config['type'] == DICT:
            result += '{} = {}\n'.format(self._output_config['root'] or 'PACKAGES_ROOT', '{')
            result += res
            result += '}\n'
        else:
            result += res

        return result

    @register_output_format('json')
    def output_format_json(self, packages, **kwargs):
        dict_or_list = self.output_format_module(packages)
        if self._output_config['root']:
            dict_or_list = {
                self._output_config['root']: dict_or_list
            }
        # TODO: Find a proper way to avoid double backslashes in path only (or not?)
        result = json.dumps(dict_or_list, indent=True)  # .replace('\\\\', '\\')
        return result



# -*- coding: utf-8 -*-
import fnmatch
import hashlib
import logging
import os
import shutil
from collections import OrderedDict

from artifactory import ArtifactoryPath

from crosspm.helpers.archive import Archive

# import only for TypeHint
try:
    from crosspm.helpers.downloader import Downloader  # noqa: F401
except ImportError:
    pass


class Package:
    def __init__(self, name, pkg, params, downloader, adapter, parser, params_found=None, params_found_raw=None,
                 stat=None, in_cache=False):
        self.name = name
        self.package_name = name
        self.packed_path = ''
        self.unpacked_path = ''
        self.duplicated = False
        self.packages = OrderedDict()

        self.pkg = pkg  # type: ArtifactoryPath
        # Someone use this internal object, do not remove  them :)
        self._pkg = self.pkg

        if isinstance(pkg, int):
            if pkg == 0:
                self._root = True

        self._raw = []
        self._root = False
        self._params_found = {}
        self._params_found_raw = {}
        self._not_cached = True
        self._log = logging.getLogger('crosspm')

        self._params = params
        self._adapter = adapter
        self._parser = parser
        self._downloader = downloader  # type: Downloader
        self._in_cache = in_cache

        if params_found:
            self._params_found = params_found
        if params_found_raw:
            self._params_found_raw = params_found_raw
        self.stat = stat

    def download(self, force=False):
        """
        Download file containing this package.
        :param force: Force download even if it seems file already exists
        :return: Full path with filename of downloaded package file.
        """
        exists, dest_path = self._downloader.cache.exists_packed(package=self, pkg_path=self.packed_path,
                                                                 check_stat=not self._in_cache)
        unp_exists, unp_path = self._downloader.cache.exists_unpacked(package=self, pkg_path=self.unpacked_path)

        # Если архива нет, то и кешу доверять не стоит
        if not exists:
            unp_exists = False

        if exists and not self.packed_path:
            self.packed_path = dest_path

        if force or not exists:
            # _packed_path = self._packed_path
            dest_path_tmp = dest_path + ".tmp"
            if os.path.exists(dest_path_tmp):
                os.remove(dest_path_tmp)
            self._adapter.download_package(self._pkg, dest_path_tmp)
            os.rename(dest_path_tmp, dest_path)
            self.packed_path = dest_path
            # if not _packed_path:
            self._not_cached = True
        else:
            if unp_exists and not self.unpacked_path:
                self.unpacked_path = unp_path
                self._not_cached = False

        if self._not_cached and unp_exists:
            shutil.rmtree(unp_path, ignore_errors=True)

        return self.packed_path

    def get_file(self, file_name, unpack_force=True):
        if unpack_force:
            self.unpack()
        _dest_file = os.path.normpath(self.get_file_path(file_name))
        _dest_file = _dest_file if os.path.isfile(_dest_file) else None
        return _dest_file

    def get_file_path(self, file_name):
        _dest_file = os.path.join(self.unpacked_path, file_name)
        return _dest_file

    def find_dependencies(self, depslock_file_path, property_validate=True, deps_content=None):
        """
        Find all dependencies by package
        :param depslock_file_path:
        :param property_validate: for `root` packages we need check property, bad if we find packages from `lock` file,
        :param deps_content: HACK for use --dependencies-content and existed dependencies.txt.lock file
        we can skip validate part
        :return:
        """
        self._raw = [x for x in
                     self._downloader.common_parser.iter_packages_params(depslock_file_path, deps_content=deps_content)]
        self.packages = self._downloader.get_dependency_packages({'raw': self._raw},
                                                                 property_validate=property_validate)

    def find_usedby(self, depslock_file_path, property_validate=True):
        """
        Find all dependencies by package
        :param depslock_file_path:
        :param property_validate: for `root` packages we need check property, bad if we find packages from `lock` file,
        we can skip validate part
        :return:
        """
        if depslock_file_path is None:
            self._raw = [self._params]
            self._raw[0]['repo'] = None
            self._raw[0]['server'] = None
        else:
            self._raw = [x for x in self._downloader.common_parser.iter_packages_params(depslock_file_path)]
        self.packages = self._downloader.get_usedby_packages({'raw': self._raw},
                                                             property_validate=property_validate)

    def unpack(self, force=False):
        if self._downloader.solid(self):
            self.unpacked_path = self.packed_path
        else:
            exists, dest_path = self._downloader.cache.exists_unpacked(package=self, pkg_path=self.unpacked_path)
            if exists and not self.unpacked_path:
                self.unpacked_path = dest_path

            # if force or not exists:

            # if not dest_path:
            #     dest_path = self._downloader.unpacked_path
            # temp_path = os.path.realpath(os.path.join(dest_path, self._name))
            # _exists = os.path.exists(temp_path)
            if not self._not_cached:
                self.unpacked_path = dest_path if exists else ''  # temp_path if exists else ''
            if force or self._not_cached or (not exists):
                Archive.extract(self.packed_path, dest_path)  # temp_path)
                self.unpacked_path = dest_path  # temp_path
                self._not_cached = False

    def pack(self, src_path):
        Archive.create(self.packed_path, src_path)

    def print(self, level=0, output=None):
        def do_print(left):
            res_str = ''
            for out_item in output:
                for k, v in out_item.items():
                    cur_str = self.get_params(merged=True).get(k, '')
                    if not res_str:
                        cur_str = self._params.get(k, '')
                    if not res_str:
                        res_str = '{}{}'.format(left, cur_str)
                        continue
                    cur_format = ' {}'
                    if v > 0:
                        cur_format = '{:%s}' % (v if len(cur_str) <= v else v + len(left))
                    res_str += cur_format.format(cur_str)
                    break
            self._log.info(res_str)

        _sign = ' '
        if not self._root:
            if self.duplicated:
                _sign = '!'
            elif self.unpacked_path:
                _sign = '+'
            elif self.packed_path:
                _sign = '>'
            else:
                _sign = '-'
        _left = '{}{}'.format(' ' * 4 * level, _sign)
        do_print(_left)
        for _pkg_name in self.packages:
            _pkg = self.packages[_pkg_name]
            if not _pkg:
                _left = '{}-'.format(' ' * 4 * (level + 1))
                self._log.info('{}{}'.format(_left, _pkg_name))
            else:
                _pkg.print(level + 1, output)
        if self._root:
            self._log.info('')

    def get_params(self, param_list=None, get_path=False, merged=False, raw=False):
        """
        Get Package params
        :param param_list: name or list of parameters
        :param get_path:
        :param merged: if version splited, True return version in string
        :param raw:
        :return:
        """
        # Convert parameter name to list
        if param_list and isinstance(param_list, str):
            param_list = [param_list]
        if param_list and isinstance(param_list, (list, tuple)):
            result = {k: v for k, v in self._params_found.items() if k in param_list}
            result.update({k: v for k, v in self._params.items() if (k in param_list and k not in result)})
        else:
            result = {k: v for k, v in self._params_found.items()}
            result.update({k: v for k, v in self._params.items() if k not in result})
        if get_path:
            result['path'] = self.unpacked_path
        if merged:
            result.update(self._parser.merge_valued(result))
        if raw:
            result.update({k: v for k, v in self._params_found_raw.items()})
        return result

    def set_full_unique_name(self):
        self.name = self._parser.get_full_package_name(self)
        return self.name

    def get_name_and_path(self, name_only=False):
        if name_only:
            return self.name
        return self.name, self.unpacked_path

    def get_none_packages(self):
        """
        Get packages with None (not founded), recursively
        """
        not_found = set()
        for package_name, package in self.packages.items():
            if package is None:
                not_found.add(package_name)
            else:
                if package.packages:
                    not_found = not_found | package.get_none_packages()
        return not_found

    @property
    def all_packages(self):
        packages = []
        for package in self.packages.values():
            if package:
                packages.extend(package.all_packages)
        packages.extend(self.packages.values())
        return packages

    def ext(self, check_ext):
        if self._pkg:
            if not isinstance(check_ext, (list, tuple)):
                check_ext = [check_ext]
            name = self._adapter.get_package_filename(self._pkg)
            if any((fnmatch.fnmatch(name, x) or fnmatch.fnmatch(name, '*%s' % x)) for x in check_ext):
                return True
        return False

    @property
    def md5(self):
        try:
            return ArtifactoryPath.stat(self.pkg).md5
        except AttributeError:
            return md5sum(self.packed_path)


def md5sum(filename):
    """
    Calculates md5 hash of a file
    """
    md5 = hashlib.md5()
    with open(filename, 'rb') as f:
        for chunk in iter(lambda: f.read(128 * md5.block_size), b''):
            md5.update(chunk)
    return md5.hexdigest()
